from __future__ import annotations
from rich import print
from rich.table import Table as RichTable
from typing import Tuple, Any, Set
from syncraft.syntax import Syntax
from syncraft.algebra import  Left, Right, Error, Either, Algebra

from syncraft.parser import ParserState, Token
from sqlglot.expressions import Expression


def rich_error(err: Error)->None:
    lst = err.to_list()
    root, leaf = lst[0], lst[-1]
    tbl = RichTable(title="Parser Error", show_lines=True)
    tbl.add_column("Root Parser Field", style="blue")
    tbl.add_column("Root Parser Value", style="green")
    tbl.add_column("...")
    tbl.add_column("Leaf Parser Field", style="blue")
    tbl.add_column("Leaf Parser Value", style="yellow")
    flds: Set[str] = set(root.keys()) | set(leaf.keys())
    for fld in sorted(flds):
        root_value = root.get(fld, "N/A")
        leaf_value = leaf.get(fld, "N/A")
        tbl.add_row(f"{fld}", f"{root_value}", "...", f"{fld}", f"{leaf_value}")
    print(tbl)


def rich_parser(p: Syntax)-> None:
    print("Parser Debug Information:")
    print(p.meta.to_string(lambda _ : True) or repr(p))

def rich_debug(this: Algebra[Any, ParserState[Any]], 
               state: ParserState[Any], 
               result: Either[Any, Tuple[Any, ParserState[Any]]])-> None:
    def value_to_str(value: Any, prefix:str='') -> str:
        if isinstance(value, (tuple, list)):
            if len(value) == 0:
                return prefix + str(value)
            else:
                return '\n'.join(value_to_str(item, prefix=prefix+' - ') for item in value)
        else:                
            if isinstance(value, Expression):
                return prefix + value.sql()
            elif isinstance(value, Token):
                return prefix + f"{value.token_type.name}({value.text})"
            elif isinstance(value, Syntax):
                return prefix + (value.meta.to_string(lambda _ : True) or 'N/A')
            else:
                return prefix + str(value)

    tbl = RichTable(title=f"Debug: {this.name}", show_lines=True)
    tbl.add_column("Parser", style="blue")
    tbl.add_column("Old State", style="cyan")
    tbl.add_column("Result", style="magenta")
    tbl.add_column("New State", style="green")
    tbl.add_column("Consumed", style="green")
    if isinstance(result, Left):
        tbl.add_row(value_to_str(this), value_to_str(state), value_to_str(result.value), 'N/A', 'N/A')
    else:
        assert isinstance(result, Right), f"Expected result to be a Right value, got {type(result)}, {result}"
        value, new_state = result.value
        tbl.add_row(value_to_str(this), 
                    value_to_str(state),
                    value_to_str(value), 
                    value_to_str(new_state), 
                    value_to_str(state.delta(new_state)))

    print(tbl)




from __future__ import annotations

from typing import (
    Optional, Any, TypeVar, Generic, Callable, Tuple, cast,
    Type, Literal
)
from dataclasses import dataclass, field, replace
from functools import reduce
from syncraft.algebra import Algebra, Error, Either, Right
from syncraft.constraint import Variable, Bindable
from syncraft.ast import Then, ThenKind, Marked
from types import MethodType, FunctionType




A = TypeVar('A')  # Result type
B = TypeVar('B')  # Result type for mapping
C = TypeVar('C')  # Result type for else branch
S = TypeVar('S', bound=Bindable)  # State type




@dataclass(frozen=True)
class Description:
    name: Optional[str] = None
    newline: Optional[str] = None
    fixity: Literal['infix', 'prefix', 'postfix'] = 'infix'
    parameter: Tuple[Any, ...] = field(default_factory=tuple)

    def update(self, 
               *,
               newline: Optional[str] = None,
               name: Optional[str] = None,
               fixity: Optional[Literal['infix', 'prefix', 'postfix']] = None,
               parameter: Optional[Tuple[Any, ...]] = None) -> 'Description':
        return Description(
            name=name if name is not None else self.name,
            newline= newline if newline is not None else self.newline,
            fixity=fixity if fixity is not None else self.fixity,
            parameter=parameter if parameter is not None else self.parameter
        )
        
    def to_string(self, interested: Callable[[Any], bool]) -> Optional[str]:
        if self.name is not None:
            if self.fixity == 'infix':
                assert len(self.parameter) == 2, "Expected exactly two parameters for infix operator"
                left  = self.parameter[0].to_string(interested) if interested(self.parameter[0]) else '...'
                right = self.parameter[1].to_string(interested) if interested(self.parameter[1]) else '...'
                if self.parameter[1].meta.newline is not None:
                    dot = '\u25cf'  
                    rarrow = '\u2794'
                    new = '\u2570' #'\u2936'
                    return f"{left}\n{new} \"{self.parameter[1].meta.newline}\" {self.name} {right}"
                return f"{left} {self.name} {right}"
            elif self.fixity == 'prefix':
                if len(self.parameter) == 0:
                    return self.name
                tmp = [x.to_string(interested) if interested(x) else '...' for x in self.parameter]
                return f"{self.name}({','.join(str(x) for x in tmp)})" 
            elif self.fixity == 'postfix':
                if len(self.parameter) == 0:
                    return self.name
                tmp = [x.to_string(interested) if interested(x) else '...' for x in self.parameter]
                return f"({','.join(str(x) for x in tmp)}).{self.name}" 
            else:
                return f"Invalid fixity: {self.fixity}"
        return None




@dataclass(frozen=True)
class Syntax(Generic[A, S]):
    alg: Callable[[Type[Algebra[Any, Any]]], Algebra[A, S]]
    meta: Description = field(default_factory=Description, repr=False)

    def algebra(self, name: str | MethodType | FunctionType, *args: Any, **kwargs: Any)-> Syntax[A, S]:
        def algebra_run(cls: Type[Algebra[Any, S]]) -> Algebra[Any, S]:
            a = self.alg(cls)
            if isinstance(name, str):
                attr = getattr(a, name, None) or getattr(cls, name, None)
                if attr is None:
                    return a
                if isinstance(attr, (staticmethod, classmethod)):
                    # These are descriptors: unwrap then call
                    attr = attr.__get__(None, cls)
                elif isinstance(attr, FunctionType):
                    # Unbound function (e.g., static method not wrapped)
                    attr = MethodType(attr, a)
                else:
                    return a
                return cast(Algebra[Any, S], attr(*args, **kwargs))
            elif isinstance(name, MethodType):
                f = MethodType(name.__func__, a)
                return cast(Algebra[Any, S], f(*args, **kwargs))
            elif isinstance(name, FunctionType):
                f = MethodType(name, a)
                return cast(Algebra[Any, S], f(*args, **kwargs))
            else:
                return a
        return self.__class__(alg=algebra_run, meta=self.meta)
                

    def as_(self, typ: Type[B])->B:
        return cast(typ, self) # type: ignore
        
    def __call__(self, alg: Type[Algebra[Any, Any]]) -> Algebra[A, S]:
        return self.alg(alg)
    
    def to_string(self, interested: Callable[[Any], bool]) -> Optional[str]:
        return self.meta.to_string(interested)

        
    def describe(self, 
                 *, 
                 newline: Optional[str] = None,
                 name: Optional[str] = None, 
                 fixity: Optional[Literal['infix', 'prefix', 'postfix']] = None, 
                 parameter: Optional[Tuple[Syntax[Any, S], ...]] = None) -> Syntax[A, S]:
        return self.__class__(alg=self.alg,
                              meta=self.meta.update(name=name,
                                    newline=newline,
                                    fixity=fixity,
                                    parameter=parameter))
    
    def newline(self, info: str='')-> Syntax[A, S]:
        return self.describe(newline=info)

    def terminal(self, name: str)->Syntax[A, S]:
        return self.describe(name=name, fixity='prefix')

######################################################## value transformation ########################################################
    def map(self, f: Callable[[A], B]) -> Syntax[B, S]:
        return self.__class__(lambda cls: self.alg(cls).map(f), meta = self.meta) # type: ignore
    
    def imap(self, f: Callable[[B], A]) -> Syntax[A, S]:
        return self.__class__(lambda cls: self.alg(cls).imap(f), meta=self.meta)
    
    def bimap(self, f: Callable[[A], B], i: Callable[[B], A]) -> Syntax[A, S]:
        return self.__class__(lambda cls: self.alg(cls).bimap(f, i), meta=self.meta)

    def map_all(self, f: Callable[[Either[Any, Tuple[A, S]]], Either[Any, Tuple[B, S]]]) -> Syntax[B, S]:
        return self.__class__(lambda cls: self.alg(cls).map_all(f), meta=self.meta) # type: ignore

    def map_error(self, f: Callable[[Optional[Any]], Any]) -> Syntax[A, S]:
        return self.__class__(lambda cls: self.alg(cls).map_error(f), meta=self.meta)
    
    def map_state(self, f: Callable[[S], S]) -> Syntax[A, S]:
        return self.__class__(lambda cls: self.alg(cls).map_state(f), meta=self.meta)
    

    def flat_map(self, f: Callable[[A], Algebra[B, S]]) -> Syntax[B, S]:
        return self.__class__(lambda cls: self.alg(cls).flat_map(f)) # type: ignore

    def many(self, *, at_least: int = 1, at_most: Optional[int] = None) -> Syntax[Tuple[A, ...], S]:
        return self.__class__(lambda cls:self.alg(cls).many(at_least=at_least, at_most=at_most)).describe(name='*', # type: ignore
                                                 fixity='prefix', 
                                                 parameter=(self,))  
    
################################################ facility combinators ############################################################



    def between(self, left: Syntax[Any, S], right: Syntax[Any, S]) -> Syntax[Then[Any, Then[A, Any]], S]:
        return left >> self // right

    def sep_by(self, sep: Syntax[Any, S]) -> Syntax[Tuple[A, ...], S]:
        return (self + (sep >> self).many().optional()).describe( # type: ignore
                    name='sep_by',
                    fixity='prefix',
                    parameter=(self, sep))
    
    def parens(self, sep: Syntax[Any, S], open: Syntax[Any, S], close: Syntax[Any, S]) -> Syntax[Any, S]:
        return self.sep_by(sep=sep).between(left=open, right=close)
            
    def optional(self, default: Optional[B] = None) -> Syntax[Optional[A | B], S]:
        return (self | success(default)).describe(name='~', fixity='prefix', parameter=(self,))


    def cut(self) -> Syntax[A, S]:
        return self.__class__(lambda cls:self.alg(cls).cut())


####################################################### operator overloading #############################################
    def __floordiv__(self, other: Syntax[B, S]) -> Syntax[Then[A, None], S]:
        other = other if isinstance(other, Syntax) else self.lift(other).as_(Syntax[B, S])
        return self.__class__(
            lambda cls: self.alg(cls).then_left(other.alg(cls))   # type: ignore
            ).describe(name=ThenKind.LEFT.value, fixity='infix', parameter=(self, other)).as_(Syntax[Then[A, None], S]) 

    def __rfloordiv__(self, other: Syntax[B, S]) -> Syntax[Then[B, None], S]:
        other = other if isinstance(other, Syntax) else self.lift(other).as_(Syntax[B, S])
        return other.__floordiv__(self)

    def __invert__(self) -> Syntax[A | None, S]:
        return self.optional()

    def __radd__(self, other: Syntax[B, S]) -> Syntax[Then[B, A], S]:
        other = other if isinstance(other, Syntax) else self.lift(other).as_(Syntax[B, S])
        return other.__add__(self)

    def __add__(self, other: Syntax[B, S]) -> Syntax[Then[A, B], S]:
        other = other if isinstance(other, Syntax) else self.lift(other).as_(Syntax[B, S])
        return self.__class__( 
            lambda cls: self.alg(cls).then_both(other.alg(cls)) # type: ignore
            ).describe(name=ThenKind.BOTH.value, fixity='infix', parameter=(self, other))

    def __rshift__(self, other: Syntax[B, S]) -> Syntax[Then[None, B], S]:
        other = other if isinstance(other, Syntax) else self.lift(other).as_(Syntax[B, S])
        return self.__class__(
            lambda cls: self.alg(cls).then_right(other.alg(cls))   # type: ignore
            ).describe(name=ThenKind.RIGHT.value, fixity='infix', parameter=(self, other)).as_(Syntax[Then[None, B], S])   


    def __rrshift__(self, other: Syntax[B, S]) -> Syntax[Then[None, A], S]:
        other = other if isinstance(other, Syntax) else self.lift(other).as_(Syntax[B, S])
        return other.__rshift__(self)  


    def __or__(self, other: Syntax[B, S]) -> Syntax[A | B, S]:
        other = other if isinstance(other, Syntax) else self.lift(other).as_(Syntax[B, S])
        return self.__class__(lambda cls: self.alg(cls).or_else(other.alg(cls))).describe(name='|', fixity='infix', parameter=[self, other]) # type: ignore


    def __ror__(self, other: Syntax[B, S]) -> Syntax[A | B, S]:
        other = other if isinstance(other, Syntax) else self.lift(other).as_(Syntax[B, S])
        return other.__or__(self).as_(Syntax[A | B, S]) 


######################################################################## data processing combinators #########################################################
    def bind(self, var: Variable) -> Syntax[A, S]:
        def bind_v(result: Either[Any, Tuple[A, S]])->Either[Any, Tuple[A, S]]:
            if isinstance(result, Right):
                value, state = result.value
                return Right((value, state.bind(var, value)))
            return result
        
        ret = self.mark(var.name).map_all(bind_v) if var.name else self.map_all(bind_v) # type: ignore
        return ret.describe(name=f'bind({var.name})', fixity='postfix', parameter=(self,))

    def mark(self, var: str) -> Syntax[A, S]:
        def bind_s(value: A) -> Marked[A]:
            if isinstance(value, Marked):
                return replace(value, name=var)    
            else:
                return Marked(name=var, value=value)
        def ibind_s(m : Marked[A]) -> A:
            return m.value if isinstance(m, Marked) else m
            
        return self.bimap(bind_s, ibind_s).describe(name=f'bind("{var}")', fixity='postfix', parameter=(self,))



    def dump_error(self, formatter: Optional[Callable[[Error], None]] = None) -> Syntax[A, S]:
        def dump_error_run(err: Any)->Any:
            if isinstance(err, Error) and formatter is not None:
                formatter(err) 
            return err
        return self.__class__(lambda cls: self.alg(cls).map_error(dump_error_run))


    def debug(self, 
              label: str, 
              formatter: Optional[Callable[[Algebra[Any, S], S, Either[Any, Tuple[Any, S]]], None]] = None) -> Syntax[A, S]:
        return self.__class__(lambda cls:self.alg(cls).debug(label, formatter), meta=self.meta)


    
def lazy(thunk: Callable[[], Syntax[A, S]]) -> Syntax[A, S]:
    return Syntax(lambda cls: cls.lazy(lambda: thunk()(cls))).describe(name='lazy(?)', fixity='postfix') 

def fail(error: Any) -> Syntax[Any, Any]:
    return Syntax(lambda alg: alg.fail(error)).describe(name=f'fail({error})', fixity='prefix')

def success(value: Any) -> Syntax[Any, Any]:
    return Syntax(lambda alg: alg.success(value)).describe(name=f'success({value})', fixity='prefix')

def choice(*parsers: Syntax[Any, S]) -> Syntax[Any, S]:
    return reduce(lambda a, b: a | b, parsers) if len(parsers) > 0 else success(None)


def all(*parsers: Syntax[Any, S]) -> Syntax[Then[Any, Any], S]:
    return reduce(lambda a, b: a + b, parsers) if len(parsers) > 0 else success(None)

def first(*parsers: Syntax[Any, S]) -> Syntax[Any, S]:
    return reduce(lambda a, b: a // b, parsers) if len(parsers) > 0 else success(None)

def last(*parsers: Syntax[Any, S]) -> Syntax[Any, S]:
    return reduce(lambda a, b: a >> b, parsers) if len(parsers) > 0 else success(None)

def bound(* parsers: Syntax[Any, S] | Tuple[str|Variable, Syntax[Any, S]]) -> Syntax[Any, S]:
    def is_named_parser(x: Any) -> bool:
        return isinstance(x, tuple) and len(x) == 2 and isinstance(x[0], (str, Variable)) and isinstance(x[1], Syntax)
    
    def to_parser(x: Syntax[Any, S] | Tuple[str|Variable, Syntax[Any, S]])->Syntax[Any, S]:
        if isinstance(x, tuple) and len(x) == 2 and isinstance(x[0], (str, Variable)) and isinstance(x[1], Syntax):
            if isinstance(x[0], str):
                return x[1].mark(x[0])
            elif isinstance(x[0], Variable):
                return x[1].bind(x[0])
            else:
                raise ValueError(f"Invalid variable type(must be str | Variable): {x[0]}", x)
        elif isinstance(x, Syntax):
            return x
        else:
            raise ValueError(f"Invalid parser or tuple: {x}", x)
    ret: Optional[Syntax[Any, S]] = None
    has_data = False
    for p in parsers:
        just_parser = to_parser(p)
        if has_data:
            if ret is not None:
                if is_named_parser(p):
                    ret = ret + just_parser
                else:
                    ret = ret // just_parser
            else:
                ret = just_parser
        else:
            has_data = is_named_parser(p)
            if ret is None:
                ret = just_parser
            else:
                ret = ret >> just_parser
    
    return ret if ret is not None else success(None) 




from __future__ import annotations

from syncraft.ast import Then, ThenKind, Many
from syncraft.algebra import Error
from syncraft.parser import literal, parse
import syncraft.generator as gen
from syncraft.generator import TokenGen
from rich import print


def test1_simple_then() -> None:
    A, B, C = literal("a"), literal("b"), literal("c")
    syntax = A // B // C
    sql = "a b c"
    ast = parse(syntax, sql, dialect="sqlite")
    print("---" * 40)
    print(ast)
    generated = gen.generate(syntax, ast)
    print("---" * 40)
    print(generated)
    assert ast == generated
    # value, bmap = generated.bimap()
    # print(value)
    # assert bmap(value) == generated


def test2_named_results() -> None:
    A, B = literal("a").mark("x").mark('z'), literal("b").mark("y")
    syntax = A // B
    sql = "a b"
    ast = parse(syntax, sql, dialect="sqlite")
    print("---" * 40)
    print(ast)
    generated = gen.generate(syntax, ast)
    print("---" * 40)
    print(generated)
    assert ast == generated
    # value, bmap = generated.bimap()
    # print(value)
    # print(bmap(value))
    # assert bmap(value) == generated


def test3_many_literals() -> None:
    A = literal("a")
    syntax = A.many()
    sql = "a a a"
    ast = parse(syntax, sql, dialect="sqlite")
    print("---" * 40)
    print(ast)
    generated = gen.generate(syntax, ast)
    print("---" * 40)
    print(generated)
    assert ast == generated
    # value, bmap = generated.bimap()
    # print(value)
    # assert bmap(value) == generated


def test4_mixed_many_named() -> None:
    A = literal("a").mark("x")
    B = literal("b")
    syntax = (A | B).many()
    sql = "a b a"
    ast = parse(syntax, sql, dialect="sqlite")
    print("---" * 40)
    print(ast)
    generated = gen.generate(syntax, ast)
    print("---" * 40)
    print(generated)
    assert ast == generated
    # value, bmap = generated.bimap()
    # print(value)
    # assert bmap(value) == generated


def test5_nested_then_many() -> None:
    IF, THEN, END = literal("if"), literal("then"), literal("end")
    syntax = (IF.many() // THEN.many()).many() // END
    sql = "if if then end"
    ast = parse(syntax, sql, dialect="sqlite")
    print("---" * 40)
    print(ast)
    generated = gen.generate(syntax, ast, restore_pruned=True)
    print("---" * 40)
    print(generated)
    assert ast == generated
    # value, bmap = generated.bimap()
    # print(value)
    # assert bmap(value) == generated



def test_then_flatten():
    A, B, C = literal("a"), literal("b"), literal("c")
    syntax = A + (B + C)
    sql = "a b c"
    ast = parse(syntax, sql, dialect='sqlite')
    print(ast)
    generated = gen.generate(syntax, ast)
    assert ast == generated
    # value, bmap = ast.bimap()
    # assert bmap(value) == ast    



def test_named_in_then():
    A = literal("a").mark("first")
    B = literal("b").mark("second")
    C = literal("c").mark("third")
    syntax = A + B + C
    sql = "a b c"
    ast = parse(syntax, sql, dialect='sqlite')
    print(ast)
    generated = gen.generate(syntax, ast)
    assert ast == generated
    # value, bmap = ast.bimap()
    # assert isinstance(value, tuple)
    # print(value)
    # assert set(x.name for x in value if isinstance(x, Marked)) == {"first", "second", "third"}
    # assert bmap(value) == ast


def test_named_in_many():
    A = literal("x").mark("x")
    syntax = A.many()
    sql = "x x x"
    ast = parse(syntax, sql, dialect='sqlite')
    print(ast)
    generated = gen.generate(syntax, ast)
    assert ast == generated
    # value, bmap = ast.bimap()
    # assert isinstance(value, list)
    # assert all(isinstance(v, Marked) for v in value if isinstance(v, Marked))
    # assert bmap(value) == ast


def test_named_in_or():
    A = literal("a").mark("a")
    B = literal("b").mark("b")
    syntax = A | B
    sql = "b"
    ast = parse(syntax, sql, dialect='sqlite')
    print(ast)
    generated = gen.generate(syntax, ast)
    assert ast == generated
    # value, bmap = ast.bimap()
    # assert isinstance(value, Marked)
    # assert value.name == "b"
    # assert bmap(value) == ast    





def test_deep_mix():
    A = literal("a").mark("a")
    B = literal("b")
    C = literal("c").mark("c")
    syntax = ((A + B) | C).many() + B
    sql = "a b a b c b"
    ast = parse(syntax, sql, dialect='sqlite')
    print(ast)
    generated = gen.generate(syntax, ast)
    print('---' * 40)
    print(generated)
    assert ast == generated
    # value, bmap = ast.bimap()
    # assert bmap(value) == ast


def test_empty_many() -> None:
    A = literal("a")
    syntax = A.many()  # This should allow empty matches
    sql = ""
    ast = parse(syntax, sql, dialect="sqlite")
    assert isinstance(ast, Error)


def test_backtracking_many() -> None:
    A = literal("a")
    B = literal("b")
    syntax = (A.many() + B)  # must not eat the final "a" needed for B
    sql = "a a a a b"
    ast = parse(syntax, sql, dialect="sqlite")
    # value, bmap = ast.bimap()
    # assert value[-1] == TokenGen.from_string("b")

def test_deep_nesting() -> None:
    A = literal("a")
    syntax = A
    for _ in range(100):
        syntax = syntax + A
    sql = " " .join("a" for _ in range(101))
    ast = parse(syntax, sql, dialect="sqlite")
    assert ast is not None


def test_nested_many() -> None:
    A = literal("a")
    syntax = (A.many().many())  # groups of groups of "a"
    sql = "a a a"
    ast = parse(syntax, sql, dialect="sqlite")
    assert isinstance(ast, Many)


def test_named_many() -> None:
    A = literal("a").mark("alpha")
    syntax = A.many()
    sql = "a a"
    ast = parse(syntax, sql, dialect="sqlite")
    # Expect [Marked("alpha", "a"), Marked("alpha", "a")]
    # flattened, _ = ast.bimap()
    # assert all(isinstance(x, Marked) for x in flattened)


def test_or_named() -> None:
    A = literal("a").mark("x")
    B = literal("b").mark("y")
    syntax = A | B
    sql = "b"
    ast = parse(syntax, sql, dialect="sqlite")
    # Either Marked("y", "b") or just "b", depending on your design
    # assert isinstance(ast, Choice)
    # value, _ = ast.bimap()
    # assert value == Marked(name="y", value=TokenGen.from_string("b"))


def test_then_associativity() -> None:
    A = literal("a")
    B = literal("b")
    C = literal("c")
    syntax = A + B + C
    sql = "a b c"
    ast = parse(syntax, sql, dialect="sqlite")
    # Should be Then(Then(A,B),C)
    assert ast == Then(kind=ThenKind.BOTH, 
                                   left=Then(kind=ThenKind.BOTH, 
                                                   left=TokenGen.from_string('a'), 
                                                   right=TokenGen.from_string('b')), 
                                    right=TokenGen.from_string('c'))


def test_ambiguous() -> None:
    A = literal("a")
    B = literal("a") + literal("b")
    syntax = A | B
    sql = "a"
    ast = parse(syntax, sql, dialect="sqlite")
    # Does it prefer A (shorter) or B (fails)? Depends on design.
    # assert ast == Choice(TokenGen.from_string("a"))


def test_combo() -> None:
    A = literal("a").mark("a")
    B = literal("b")
    C = literal("c").mark("c")
    syntax = ((A + B).many() | C) + B
    sql = "a b a b c b"
    # Should fail, as we discussed earlier
    ast = parse(syntax, sql, dialect="sqlite")
    assert isinstance(ast, Error)


def test_optional():
    A = literal("a").mark("a")
    syntax = A.optional()
    ast1 = parse(syntax, "", dialect="sqlite")
    # v1, _ = ast1.bimap()
    # assert v1 is None
    # ast2 = parse(syntax, "a", dialect="sqlite")
    # v2, _ = ast2.bimap()
    # assert v2 == Marked(name='a', value=TokenGen.from_string('a'))








from __future__ import annotations
from typing import (
    Optional, List, Any, TypeVar, Generic, Callable, Tuple, cast, 
    Dict, Type, ClassVar, Hashable
)

import traceback
from dataclasses import dataclass, replace, asdict
from weakref import WeakKeyDictionary
from abc import ABC
from syncraft.ast import ThenKind, Then, S, Choice, Many, ChoiceKind
    
A = TypeVar('A')  # Result type
B = TypeVar('B')  # Mapped result type


InProgress = object()  # Marker for in-progress state, used to prevent re-entrance in recursive calls
L = TypeVar('L')  # Left type for combined results
R = TypeVar('R')  # Right type for combined results

class Either(Generic[L, R]):
    def is_left(self) -> bool:
        return isinstance(self, Left)
    def is_right(self) -> bool:
        return isinstance(self, Right)

@dataclass(frozen=True)
class Left(Either[L, R]):
    value: Optional[L] = None

@dataclass(frozen=True)
class Right(Either[L, R]):
    value: R




@dataclass(frozen=True)
class Error:
    this: Any
    message: Optional[str] = None
    error: Optional[Any] = None    
    state: Optional[Any] = None
    committed: bool = False
    previous: Optional[Error] = None
    
    def attach( self, 
                *,
                this: Any, 
                msg: Optional[str] = None,
                err: Optional[str] = None, 
                state: Optional[Any] = None) -> Error:
        return Error(
            this=this,
            error=err,
            message=msg or str(err),
            state=state,
            previous=self
        )
    def to_list(self)->List[Dict[str, Any]]:
        lst = []
        current: Optional[Error] = self
        while current is not None:
            d = asdict(current)
            lst.append({k:v for k,v in d.items() if v is not None and k != 'previous'})
            current = current.previous
        return lst


@dataclass(frozen=True)        
class Algebra(ABC, Generic[A, S]):
######################################################## shared among all subclasses ########################################################
    run_f: Callable[[S, bool], Either[Any, Tuple[A, S]]] 
    name: Hashable
    _cache: ClassVar[WeakKeyDictionary[Any, Dict[Any, object | Either[Any, Tuple[Any, Any]]]]] = WeakKeyDictionary()

    def named(self, name: Hashable) -> 'Algebra[A, S]':
        return replace(self, name=name)

    def __post_init__(self)-> None:
        self._cache.setdefault(self.run_f, dict())
        
    def __call__(self, input: S, use_cache: bool) -> Either[Any, Tuple[A, S]]:
        return self.run(input, use_cache=use_cache)

    
    def run(self, input: S, use_cache: bool) -> Either[Any, Tuple[A, S]]:
        cache = self._cache[self.run_f]
        assert cache is not None, "Cache should be initialized in __post_init__"
        if input in cache:
            v = cache.get(input, None)
            if v is InProgress:
                return Left(
                    Error(
                        message="Left-recursion detected in parser",
                        this=self,
                        state=input
                    ))
            else:
                return cast(Either[Error, Tuple[A, S]], v)
        try:
            cache[input] = InProgress
            result = self.run_f(input, use_cache)
            cache[input] = result
            if not use_cache:
                cache.pop(input, None)  # Clear the cache entry if not using cache
            if isinstance(result, Left):
                if isinstance(result.value, Error):
                    result = Left(result.value.attach(this=self, state=input))
        except Exception as e:
            cache.pop(input, None)  # Clear the cache entry on exception
            traceback.print_exc()
            print(f"Exception from self.run(S): {e}")
            return Left(
                Error(
                    message="Exception from self.run(S): {e}",
                    this=self,
                    state=input,
                    error=e
                ))
        return result

    def as_(self, typ: Type[B])->B:
        return cast(typ, self) # type: ignore
        
    @classmethod
    def lazy(cls, thunk: Callable[[], Algebra[A, S]]) -> Algebra[A, S]:
        def lazy_run(input: S, use_cache:bool) -> Either[Any, Tuple[A, S]]:
            return thunk().run(input, use_cache)
        return cls(lazy_run, name=cls.__name__ + '.lazy')




    @classmethod
    def fail(cls, error: Any) -> Algebra[Any, S]:
        def fail_run(input: S, use_cache:bool) -> Either[Any, Tuple[Any, S]]:
            return Left(Error(
                error=error,
                this=cls,
                state=input
            ))
        return cls(fail_run, name=cls.__name__ + '.fail')
    @classmethod
    def success(cls, value: Any) -> Algebra[Any, S]:
        def success_run(input: S, use_cache:bool) -> Either[Any, Tuple[Any, S]]:
            return Right((value, input))
        return cls(success_run, name=cls.__name__ + '.success')
    
    @classmethod
    def factory(cls, name: str, *args: Any, **kwargs: Any) -> Algebra[A, S]:
        method = getattr(cls, name, None)
        if method is None or not callable(method):
            raise ValueError(f"Method {name} is not defined in {cls.__name__}")
        return cast(Algebra[A, S], method(*args, **kwargs))



    def cut(self) -> Algebra[A, S]:
        def commit_error(e: Any) -> Error:
            match e:
                case Error():
                    return replace(e, committed=True)
                case _:
                    return Error(
                        error=e,
                        this=self,
                        committed=True
                    )
        return self.map_error(commit_error)

    def on_fail(self, 
                func: Callable[
                    [
                        Algebra[A, S], 
                        S, 
                        Left[Any, Tuple[A, S]], 
                        Any
                    ], 
                    Either[Any, Tuple[B, S]]], 
                    ctx: Optional[Any] = None) -> Algebra[A | B, S]:
        assert callable(func), "func must be callable"
        def fail_run(input: S, use_cache:bool) -> Either[Any, Tuple[A | B, S]]:
            result = self.run(input, use_cache)
            if isinstance(result, Left):
                return cast(Either[Any, Tuple[A | B, S]], func(self, input, result, ctx))
            return cast(Either[Any, Tuple[A | B, S]], result)
        return self.__class__(fail_run, name=self.name) # type: ignore

    def on_success(self, 
                    func: Callable[
                        [
                            Algebra[A, S], 
                            S, 
                            Right[Any, Tuple[A, S]], 
                            Any
                        ], 
                        Either[Any, Tuple[B, S]]], 
                        ctx: Optional[Any] = None) -> Algebra[A | B, S]:
        assert callable(func), "func must be callable"
        def success_run(input: S, use_cache:bool) -> Either[Any, Tuple[A | B, S]]:
            result = self.run(input, use_cache)
            if isinstance(result, Right):
                return cast(Either[Any, Tuple[A | B, S]], func(self, input, result, ctx))
            return cast(Either[Any, Tuple[A | B, S]], result)
        return self.__class__(success_run, name=self.name) # type: ignore

    def debug(self, 
              label: str, 
              formatter: Optional[Callable[[
                  Algebra[Any, S], 
                  S, 
                  Either[Any, Tuple[Any, S]]], None]]=None) -> Algebra[A, S]:
        def default_formatter(alg: Algebra[Any, S], input: S, result: Either[Any, Tuple[Any, S]]) -> None:
            print(f"Debug: {'*' * 40} {alg.name} - State {'*' * 40}")
            print(input)
            print(f"Debug: {'~' * 40} (Result, State) {'~' * 40}")
            print(result)
            print()
            print()
        lazy_self: Algebra[A, S]
        def debug_run(input: S, use_cache:bool) -> Either[Any, Tuple[A, S]]:
            result = self.run(input, use_cache)
            try:
                if formatter is not None:
                    formatter(lazy_self, input, result)
                else:
                    default_formatter(lazy_self, input, result)
            except Exception as e:
                traceback.print_exc()
                print(f"Error occurred while formatting debug information: {e}")
            finally:
                return result
        lazy_self = self.__class__(debug_run, name=label)  
        return lazy_self


######################################################## fundamental combinators ############################################    
    def fmap(self, f: Callable[[A], B]) -> Algebra[B, S]:
        def fmap_run(input: S, use_cache:bool) -> Either[Any, Tuple[B, S]]:
            parsed = self.run(input, use_cache)
            if isinstance(parsed, Right):
                return Right((f(parsed.value[0]), parsed.value[1]))            
            else:
                return cast(Either[Any, Tuple[B, S]], parsed)
        return self.__class__(fmap_run, name=self.name)  # type: ignore

    def imap(self, f: Callable[[B], A]) -> Algebra[A, S]:
        return self.map_state(lambda s: s.map(f))
    
    def map(self, f: Callable[[A], B]) -> Algebra[B, S]:
        return self.fmap(f)
    
    def bimap(self, f: Callable[[A], B], i: Callable[[B], A]) -> Algebra[A, S]:
        return self.fmap(f).as_(Algebra[A, S]).imap(i)

    def map_all(self, f: Callable[[Either[Any, Tuple[A, S]]], Either[Any, Tuple[B, S]]])->Algebra[B, S]:
        def map_all_run(input: S, use_cache:bool) -> Either[Any, Tuple[B, S]]:
            parsed = self.run(input, use_cache)
            return f(parsed)
        return self.__class__(map_all_run, name=self.name) # type: ignore

    def map_error(self, f: Callable[[Optional[Any]], Any]) -> Algebra[A, S]:
        def map_error_run(input: S, use_cache:bool) -> Either[Any, Tuple[A, S]]:
            parsed = self.run(input, use_cache)
            if isinstance(parsed, Left):
                return Left(f(parsed.value))
            return parsed
        return self.__class__(map_error_run, name=self.name)  

    def map_state(self, f: Callable[[S], S]) -> Algebra[A, S]:
        def map_state_run(state: S, use_cache:bool) -> Either[Any, Tuple[A, S]]:
            return self.run(f(state), use_cache)
        return self.__class__(map_state_run, name=self.name) 


    def flat_map(self, f: Callable[[A], Algebra[B, S]]) -> Algebra[B, S]:
        def flat_map_run(input: S, use_cache:bool) -> Either[Any, Tuple[B, S]]:
            parsed = self.run(input, use_cache)
            if isinstance(parsed, Right):
                return f(parsed.value[0]).run(parsed.value[1], use_cache)  
            else:
                return cast(Either[Any, Tuple[B, S]], parsed)
        return self.__class__(flat_map_run, name=self.name)  # type: ignore

    
    def or_else(self: Algebra[A, S], other: Algebra[B, S]) -> Algebra[Choice[A, B], S]:
        def or_else_run(input: S, use_cache:bool) -> Either[Any, Tuple[Choice[A, B], S]]:
            match self.run(input, use_cache):
                case Right((value, state)):
                    return Right((Choice(kind=ChoiceKind.LEFT, left=value, right=None), state))
                case Left(err):
                    if isinstance(err, Error) and err.committed:
                        return Left(err)
                    match other.run(input, use_cache):
                        case Right((other_value, other_state)):
                            return Right((Choice(kind=ChoiceKind.RIGHT, left=None, right=other_value), other_state))
                        case Left(other_err):
                            return Left(other_err)
                    raise TypeError(f"Unexpected result type from {other}")
            raise TypeError(f"Unexpected result type from {self}")
        return self.__class__(or_else_run, name=f'{self.name} | {other.name}')  # type: ignore

    def then_both(self, other: 'Algebra[B, S]') -> 'Algebra[Then[A, B], S]':
        def then_both_f(a: A) -> Algebra[Then[A, B], S]:
            def combine(b: B) -> Then[A, B]:
                return Then(left=a, right=b, kind=ThenKind.BOTH)
            return other.fmap(combine)
        return self.flat_map(then_both_f).named(f'{self.name} + {other.name}')

    def then_left(self, other: Algebra[B, S]) -> 'Algebra[Then[A, B], S]':
        def then_left_f(a: A) -> Algebra[Then[A, B], S]:
            def combine(b: B) -> Then[A, B]:
                return Then(left=a, right=b, kind=ThenKind.LEFT)
            return other.fmap(combine)
        return self.flat_map(then_left_f).named(f'{self.name} // {other.name}')

    def then_right(self, other: Algebra[B, S]) -> 'Algebra[Then[A, B], S]':
        def then_right_f(a: A) -> Algebra[Then[A, B], S]:
            def combine(b: B) -> Then[A, B]:
                return Then(left=a, right=b, kind=ThenKind.RIGHT)
            return other.fmap(combine)
        return self.flat_map(then_right_f).named(f'{self.name} >> {other.name}')

    def many(self, *, at_least: int, at_most: Optional[int]) -> Algebra[Many[A], S]:
        assert at_least > 0, "at_least must be greater than 0"
        assert at_most is None or at_least <= at_most, "at_least must be less than or equal to at_most"
        def many_run(input: S, use_cache:bool) -> Either[Any, Tuple[Many[A], S]]:
            ret: List[A] = []
            current_input = input
            while True:
                match self.run(current_input, use_cache):
                    case Left(_):
                        break
                    case Right((value, next_input)):
                        ret.append(value)
                        if next_input == current_input:
                            break  # No progress, stop to avoid infinite loop
                        current_input = next_input
                        if at_most is not None and len(ret) > at_most:
                            return Left(Error(
                                    message=f"Expected at most {at_most} matches, got {len(ret)}",
                                    this=self,
                                    state=current_input
                                )) 
            if len(ret) < at_least:
                return Left(Error(
                        message=f"Expected at least {at_least} matches, got {len(ret)}",
                        this=self,
                        state=current_input
                    )) 
            return Right((Many(value=tuple(ret)), current_input))
        return self.__class__(many_run, name=f'*({self.name})') # type: ignore

    





from typing import Any
from syncraft.parser import parse, until, literal, Parser
from syncraft.ast import AST

# Define common pair DSLs
LP, RP = literal("("), literal(")")
LB, RB = literal("["), literal("]")


def test_until_accepts_proper_nesting() -> None:
    sql = "([])"
    syntax = until((LP, RP), (LB, RB))
    ast: AST | Any = parse(syntax, sql, dialect="sqlite")
    assert isinstance(ast, tuple), f"Expected AST for proper nesting, got {ast}"


def test_until_rejects_mismatched_pairs() -> None:
    # Mismatched: ( ] should fail immediately
    sql = "(]"
    syntax = until((LP, RP), (LB, RB))
    res = parse(syntax, sql, dialect="sqlite")
    from syncraft.algebra import Error
    assert isinstance(res, Error), "Mismatched pairs should be rejected with an Error"

def test_until_rejects_unterminated_group() -> None:
    # Unterminated: ( ... EOF
    sql = "("
    syntax = until((LP, RP))
    res = parse(syntax, sql, dialect="sqlite")
    from syncraft.algebra import Error
    assert isinstance(res, Error), "Unterminated group should be rejected with an Error"

def test_until_rejects_crossing_pairs() -> None:
    # Crossing/interleaved: ([)] should be rejected
    sql = "([)]"
    syntax = until((LP, RP), (LB, RB))
    # Use postgres dialect so [ and ] are tokenized distinctly (not as bracketed identifier)
    res = parse(syntax, sql, dialect="postgres")
    from syncraft.algebra import Error
    assert isinstance(res, Error), "Crossing pairs should be rejected with an Error"



from __future__ import annotations

from typing import (
    Any, TypeVar, Tuple, Optional,  Callable, Generic, 
    List, 
)
from functools import cached_property
from dataclasses import dataclass, replace
from syncraft.algebra import (
    Algebra, Either, Left, Right, Error
)

from syncraft.ast import (
    T, ParseResult, AST, Token, TokenSpec, 
    Bindable, 
    Choice, Many, ChoiceKind,
    Then, ThenKind, Marked
)

from syncraft.syntax import Syntax
from sqlglot import TokenType
import re
import rstr
from functools import lru_cache
import random

B = TypeVar('B')


@dataclass(frozen=True)
class GenState(Bindable, Generic[T]):
    ast: Optional[ParseResult[T]] = None
    restore_pruned: bool = False
    seed: int = 0
    def map(self, f: Callable[[Any], Any]) -> GenState[T]:
        return replace(self, ast=f(self.ast))
    
    def inject(self, a: Any) -> GenState[T]:
        return self.map(lambda _: a)
    
    def fork(self, tag: Any) -> GenState[T]:
        return replace(self, seed=hash((self.seed, tag)))

    def rng(self, tag: Any = None) -> random.Random:
        return random.Random(self.seed if tag is None else hash((self.seed, tag)))



    @cached_property
    def pruned(self)->bool:
        return self.ast is None
    
    def left(self)-> GenState[T]:
        if self.ast is None:
            return self
        if isinstance(self.ast, Then) and (self.ast.kind != ThenKind.RIGHT or self.restore_pruned):
            return replace(self, ast=self.ast.left)
        return replace(self, ast=None) 

    def right(self) -> GenState[T]:
        if self.ast is None:
            return self
        if isinstance(self.ast, Then) and (self.ast.kind != ThenKind.LEFT or self.restore_pruned):
            return replace(self, ast=self.ast.right)
        return replace(self, ast=None)
    

    
    def down(self, index: int) -> GenState[T]:
        if self.ast is None:
            return self        
        match self.ast:
            case Marked(value=value):
                return replace(self, ast=value)
            case _:
                raise TypeError(f"Invalid AST type({self.ast}) for down traversal")
    
    @classmethod
    def from_ast(cls, *, ast: Optional[ParseResult[T]], seed: int = 0, restore_pruned:bool=False) -> GenState[T]:
        return cls(ast=ast, seed=seed, restore_pruned=restore_pruned)
    




@lru_cache(maxsize=None)
def token_type_from_string(token_type: Optional[TokenType], text: str, case_sensitive:bool)-> TokenType:
    if not isinstance(token_type, TokenType) or token_type == TokenType.VAR:
        for t in TokenType:
            if t.value == text or str(t.value).lower() == text.lower():
                return t
        return TokenType.VAR
    return token_type


@dataclass(frozen=True)
class TokenGen(TokenSpec):

    def __str__(self) -> str:
        tt = self.token_type.name if self.token_type else ""
        txt = self.text if self.text else ""
        reg = self.regex.pattern if self.regex else ""
        return f"TokenGen({tt}, {txt}, {self.case_sensitive}, {reg})"
        
    
    def __repr__(self) -> str:
        return self.__str__()

    def gen(self) -> Token:
        text: str
        if self.text is not None:
            text = self.text
        elif self.regex is not None:
            try:
                text = rstr.xeger(self.regex)
            except Exception as e:
                # If the regex is invalid or generation fails
                text = self.regex.pattern  # fallback to pattern string
        elif self.token_type is not None:
            text = str(self.token_type.value)
        else:
            text = "VALUE"

        return Token(token_type= token_type_from_string(self.token_type,
                                                        text, 
                                                        self.case_sensitive), 
                     text=text)        

    @staticmethod
    def from_string(string: str)->Token:
        return Token(token_type=token_type_from_string(None, string, case_sensitive=False), text=string)


@dataclass(frozen=True)
class Generator(Algebra[ParseResult[T], GenState[T]]):  
    def flat_map(self, f: Callable[[ParseResult[T]], Algebra[B, GenState[T]]]) -> Algebra[B, GenState[T]]: 
        def flat_map_run(input: GenState[T], use_cache:bool) -> Either[Any, Tuple[B, GenState[T]]]:
            try:
                if not isinstance(input.ast, Then):
                    return Left(Error(this=self, 
                                      message=f"Expect Then got {input.ast}",
                                      state=input))
                lft = input.left() 
                match self.run(lft, use_cache=use_cache):
                    case Left(error):
                        return Left(error)
                    case Right((value, next_input)):
                        r = input.right() 
                        match f(value).run(r, use_cache):
                            case Left(e):
                                return Left(e)
                            case Right((result, next_input)):
                                return Right((result, next_input))
                raise ValueError("flat_map should always return a value or an error.")
            except Exception as e:
                return Left(Error(
                    message=str(e),
                    this=self,
                    state=input,
                    error=e
                ))
        return self.__class__(run_f = flat_map_run, name=self.name) # type: ignore


    def many(self, *, at_least: int, at_most: Optional[int]) -> Algebra[Many[ParseResult[T]], GenState[T]]:
        assert at_least > 0, "at_least must be greater than 0"
        assert at_most is None or at_least <= at_most, "at_least must be less than or equal to at_most"
        def many_run(input: GenState[T], use_cache:bool) -> Either[Any, Tuple[Many[ParseResult[T]], GenState[T]]]:
            if input.pruned:
                upper = at_most if at_most is not None else at_least + 2
                count = input.rng("many").randint(at_least, upper)
                ret: List[Any] = []
                for i in range(count):
                    forked_input = input.fork(tag=len(ret))
                    match self.run(forked_input, use_cache):
                        case Right((value, next_input)):
                            ret.append(value)
                        case Left(_):
                            pass
                return Right((Many(value=tuple(ret)), input))
            else:
                if not isinstance(input.ast, Many):
                    return Left(Error(this=self, 
                                      message=f"Expect Many got {input.ast}",
                                      state=input))
                ret = []
                for x in input.ast.value: 
                    match self.run(input.inject(x), use_cache):
                        case Right((value, next_input)):
                            ret.append(value)
                            if at_most is not None and len(ret) > at_most:
                                return Left(Error(
                                        message=f"Expected at most {at_most} matches, got {len(ret)}",
                                        this=self,
                                        state=input.inject(x)
                                    ))                             
                        case Left(_):
                            pass
                if len(ret) < at_least:
                    return Left(Error(
                        message=f"Expected at least {at_least} matches, got {len(ret)}",
                        this=self,
                        state=input.inject(x)
                    )) 
                return Right((Many(value=tuple(ret)), input))
        return self.__class__(many_run, name=f"many({self.name})")  # type: ignore
    
 
    def or_else(self, # type: ignore
                other: Algebra[ParseResult[T], GenState[T]]
                ) -> Algebra[Choice[ParseResult[T], ParseResult[T]], GenState[T]]: 
        def or_else_run(input: GenState[T], use_cache:bool) -> Either[Any, Tuple[Choice[ParseResult[T], ParseResult[T]], GenState[T]]]:
            def exec(kind: ChoiceKind, 
                     left: GenState[T], 
                     right: GenState[T])->Either[Any, Tuple[Choice[ParseResult[T], ParseResult[T]], GenState[T]]]:
                match kind:
                    case ChoiceKind.LEFT:
                        match self.run(left, use_cache):
                            case Right((value, next_input)):
                                return Right((Choice(kind=ChoiceKind.LEFT, left=value, right=None), next_input))
                            case Left(error):
                                return Left(error)
                    case ChoiceKind.RIGHT:
                        match other.run(right, use_cache):
                            case Right((value, next_input)):
                                return Right((Choice(kind=ChoiceKind.RIGHT, left=None, right=value), next_input))
                            case Left(error):
                                return Left(error)
                raise ValueError(f"Invalid ChoiceKind: {kind}")

            if input.pruned:
                forked_input = input.fork(tag="or_else")
                which = forked_input.rng("or_else").choice((ChoiceKind.LEFT, ChoiceKind.RIGHT))
                return exec(which, forked_input, forked_input)
            else:
                if isinstance(input.ast, Choice):
                    return exec(input.ast.kind, 
                                input.inject(input.ast.left), 
                                input.inject(input.ast.right))
                else:
                    return Left(Error(this=self, 
                                      message=f"Expect Choice got {input.ast}",
                                      state=input))
        return self.__class__(or_else_run, name=f"or_else({self.name} | {other.name})") # type: ignore

    @classmethod
    def token(cls, 
              token_type: Optional[TokenType] = None, 
              text: Optional[str] = None, 
              case_sensitive: bool = False,
              regex: Optional[re.Pattern[str]] = None
              )-> Algebra[ParseResult[T], GenState[T]]:      
        gen = TokenGen(token_type=token_type, text=text, case_sensitive=case_sensitive, regex=regex)  
        lazy_self: Algebra[ParseResult[T], GenState[T]]
        def token_run(input: GenState[T], use_cache:bool) -> Either[Any, Tuple[ParseResult[Token], GenState[T]]]:
            if input.pruned:
                return Right((gen.gen(), input))
            else:
                current = input.ast
                if not isinstance(current, Token) or not gen.is_valid(current):
                    return Left(Error(None, 
                                      message=f"Expected a Token, but got {type(current)}.", 
                                      state=input))
                return Right((current, input))
        lazy_self = cls(token_run, name=cls.__name__ + f'.token({token_type or text or regex})')  # type: ignore
        return lazy_self



def generate(syntax: Syntax[Any, Any], 
            data: Optional[ParseResult[Any]] = None, 
            seed: int = 0, 
            restore_pruned: bool = False) -> AST | Any:
    gen = syntax(Generator)
    state = GenState.from_ast(ast=data, seed=seed, restore_pruned=restore_pruned)
    result = gen.run(state, use_cache=False)
    if isinstance(result, Right):
        return result.value[0]
    assert isinstance(result, Left), "Generator must return Either[Any, Tuple[Any, Any]]"
    return result.value


    


from syncraft.ast import AST
from syncraft.parser import literal, variable, parse
import syncraft.generator as gen
from typing import Any

IF = literal("if")
ELSE = literal("else")
THEN = literal("then")
END = literal("end")
var = variable()


def test_between()->None:
    sql = "then if then"
    syntax = IF.between(THEN, THEN)
    ast:AST = parse(syntax, sql, dialect='sqlite')    
    generated = gen.generate(syntax, ast)
    assert ast == generated, "Parsed and generated results do not match."


def test_sep_by()->None:
    sql = "if then if then if then if"
    syntax = IF.sep_by(THEN)
    ast:AST = parse(syntax, sql, dialect='sqlite')    
    generated = gen.generate(syntax, ast)
    assert ast == generated, "Parsed and generated results do not match."

def test_many_or()->None:
    IF = literal("if")
    THEN = literal("then")
    END = literal("end")
    syntax = (IF.many() | THEN.many()).many() // END
    sql = "if if then end"
    ast:AST = parse(syntax, sql, dialect='sqlite')
    generated = gen.generate(syntax, ast)
    assert ast == generated, "Parsed and generated results do not match."



from __future__ import annotations

from typing import (
    Any, Tuple, Generator as YieldGen
)
from dataclasses import dataclass
from syncraft.algebra import (
    Algebra, Either, Right, 
)
from syncraft.ast import T, ParseResult, Choice, Many, Then, Marked

from syncraft.generator import GenState, Generator

from syncraft.syntax import Syntax



@dataclass(frozen=True)
class Finder(Generator[T]):
    @classmethod
    def anything(cls)->Algebra[Any, GenState[T]]:
        def anything_run(input: GenState[T], use_cache:bool) -> Either[Any, Tuple[Any, GenState[T]]]:
            return Right((input.ast, input))
        return cls(anything_run, name=cls.__name__ + '.anything()')



anything = Syntax(lambda cls: cls.factory('anything')).describe(name="anything", fixity='infix') 

def matches(syntax: Syntax[Any, Any], data: ParseResult[T])-> bool:
    gen = syntax(Finder)
    state = GenState.from_ast(ast = data, restore_pruned=True)
    result = gen.run(state, use_cache=True)
    return isinstance(result, Right)


def find(syntax: Syntax[Any, Any], data: ParseResult[T]) -> YieldGen[ParseResult[T], None, None]:
    if matches(syntax, data):
        yield data
    match data:
        case Then(left=left, right=right):
            if left is not None:
                yield from find(syntax, left)
            if right is not None:
                yield from find(syntax, right)
        case Many(value = value):
            for e in value:
                yield from find(syntax, e)
        case Marked(value=value):
            yield from find(syntax, value)
        case Choice(left=left, right=right):
            if left is not None:
                yield from find(syntax, left)
            if right is not None:
                yield from find(syntax, right)
        case _:
            pass



from __future__ import annotations
import re
from sqlglot import tokenize, TokenType, Parser as GlotParser, exp
from typing import (
    Optional, List, Any, Tuple,
    Generic
)
from syncraft.algebra import (
    Either, Left, Right, Error, Algebra
)
from dataclasses import dataclass, field, replace
from enum import Enum
from functools import reduce
from syncraft.syntax import Syntax

from syncraft.ast import Token, TokenSpec, AST, T, Bindable


@dataclass(frozen=True)
class ParserState(Bindable, Generic[T]):
    input: Tuple[T, ...] = field(default_factory=tuple)
    index: int = 0

    
    def token_sample_string(self)-> str:
        def encode_tokens(*tokens:T) -> str:
            return ",".join(f"{token.token_type.name}({token.text})" for token in tokens)
        return encode_tokens(*self.input[self.index:self.index + 2])

    def before(self, length: Optional[int] = 5)->str:
        length = min(self.index, length) if length is not None else self.index
        return " ".join(token.text for token in self.input[self.index - length:self.index])
    
    def after(self, length: Optional[int] = 5)->str:
        length = min(length, len(self.input) - self.index) if length is not None else len(self.input) - self.index
        return " ".join(token.text for token in self.input[self.index:self.index + length])


    def current(self)->T:
        if self.ended():
            raise IndexError("Attempted to access token beyond end of stream")
        return self.input[self.index]
    
    def ended(self) -> bool:
        return self.index >= len(self.input)

    def advance(self) -> ParserState[T]:
        return replace(self, index=min(self.index + 1, len(self.input)))
            
    def delta(self, new_state: ParserState[T]) -> Tuple[T, ...]:
        assert self.input is new_state.input, "Cannot calculate differences between different input streams"
        assert 0 <= self.index <= new_state.index <= len(self.input), "Segment indices out of bounds"
        return self.input[self.index:new_state.index]
    
    def copy(self) -> ParserState[T]:
        return self.__class__(input=self.input, index=self.index)

    @classmethod
    def from_tokens(cls, tokens: Tuple[T, ...]) -> ParserState[T]:
        return cls(input=tokens, index=0)




    
@dataclass(frozen=True)
class Parser(Algebra[T, ParserState[T]]):
    @classmethod
    def token(cls, 
              token_type: Optional[Enum] = None, 
              text: Optional[str] = None, 
              case_sensitive: bool = False,
              regex: Optional[re.Pattern[str]] = None
              )-> Algebra[T, ParserState[T]]:
        spec = TokenSpec(token_type=token_type, text=text, case_sensitive=case_sensitive, regex=regex)
        def token_run(state: ParserState[T], use_cache:bool) -> Either[Any, Tuple[T, ParserState[T]]]:
            if state.ended():
                return Left(state)
            token = state.current()
            if token is None or not spec.is_valid(token):
                return Left(state)
            return Right((Token(token_type = token.token_type, text=token.text), state.advance()))  # type: ignore
        captured: Algebra[T, ParserState[T]] = cls(token_run, name=cls.__name__ + f'.token({token_type}, {text})')
        def error_fn(err: Any) -> Error:
            if isinstance(err, ParserState):
                return Error(message=f"Cannot match token at {err}", this=captured, state=err)            
            else:
                return Error(message="Cannot match token at unknown state", this=captured)
        # assign the updated parser(with description) to bound variable so the Error.this could be set correctly
        captured = captured.map_error(error_fn)
        return captured        


    @classmethod
    def until(cls, 
              *open_close: Tuple[Algebra[Any, ParserState[T]], Algebra[Any, ParserState[T]]],
              terminator: Optional[Algebra[Any, ParserState[T]]] = None,
              inclusive: bool = True, 
              strict: bool = True) -> Algebra[Any, ParserState[T]]:
        def until_run(state: ParserState[T], use_cache:bool) -> Either[Any, Tuple[Any, ParserState[T]]]:
            # Use a stack to enforce proper nesting across multiple open/close pairs.
            tokens: List[Any] = []
            if not terminator and len(open_close) == 0:
                return Left(Error(this=until_run, message="No terminator and no open/close parsers, nothing to parse", state=state))

            # Helper to try matching any of the parsers once, returning early on first match
            def try_match(s: ParserState[T], *parsers: Algebra[Any, ParserState[T]]) -> Tuple[bool, Optional[int], Optional[Any], ParserState[T]]:
                for i, p in enumerate(parsers):
                    res = p.run(s, use_cache)
                    if isinstance(res, Right):
                        val, ns = res.value
                        return True, i, val, ns
                return False, None, None, s

            opens, closes = zip(*open_close) if len(open_close) > 0 else ((), ())
            tmp_state: ParserState[T] = state.copy()
            stack: List[int] = []  # indices into open_close indicating expected closer

            # If strict, require the very next token to be an opener of any kind
            if strict and len(opens) > 0:
                c = reduce(lambda a, b: a.or_else(b), opens).run(tmp_state, use_cache)
                if c.is_left():
                    return Left(Error(this=until_run, message="No opening parser matched", state=tmp_state))

            while not tmp_state.ended():
                # Try to open
                o_matched, o_idx, o_tok, o_state = try_match(tmp_state, *opens)
                if o_matched and o_idx is not None:
                    stack.append(o_idx)
                    if inclusive:
                        tokens.append(o_tok)
                    tmp_state = o_state
                    continue

                # Try to close
                c_matched, c_idx, c_tok, c_state = try_match(tmp_state, *closes)
                if c_matched and c_idx is not None:
                    if not stack or stack[-1] != c_idx:
                        return Left(Error(this=until_run, message="Mismatched closing parser", state=tmp_state))
                    stack.pop()
                    if inclusive:
                        tokens.append(c_tok)
                    tmp_state = c_state
                    # After closing, if stack empty, we may terminate on a terminator
                    if len(stack) == 0:
                        if terminator:
                            term = terminator.run(tmp_state, use_cache)
                            if isinstance(term, Right):
                                if inclusive:
                                    tokens.append(term.value[0])
                                return Right((tuple(tokens), term.value[1]))
                        else:
                            return Right((tuple(tokens), tmp_state))
                    continue

                # If nothing structural matched, check termination when not nested
                if len(stack) == 0:
                    if terminator:
                        term2 = terminator.run(tmp_state, use_cache)
                        if isinstance(term2, Right):
                            if inclusive:
                                tokens.append(term2.value[0])
                            return Right((tuple(tokens), term2.value[1]))
                    else:
                        return Right((tuple(tokens), tmp_state))

                # Otherwise, consume one token as payload and continue
                tokens.append(tmp_state.current())
                tmp_state = tmp_state.advance()

            # Reached end of input
            if len(stack) != 0:
                return Left(Error(this=until_run, message="Unterminated group", state=tmp_state))
            return Right((tuple(tokens), tmp_state))
        return cls(until_run, name=cls.__name__ + '.until')

def sqlglot(parser: Syntax[Any, Any], 
            dialect: str) -> Syntax[List[exp.Expression], ParserState[Any]]:
    gp = GlotParser(dialect=dialect)
    return parser.map(lambda tokens: [e for e in gp.parse(raw_tokens=tokens) if e is not None])


def parse(syntax: Syntax[Any, Any], 
          sql: str, 
          dialect: str) -> AST | Any:
    parser = syntax(Parser)
    input: ParserState[Token] = token_state(sql, dialect=dialect)
    result = parser.run(input, True)
    if isinstance(result, Right):
        return result.value[0]
    assert isinstance(result, Left), "Parser must return Either[E, Tuple[A, S]]"
    return result.value


def token_state(sql: str, dialect: str) -> ParserState[Token]:
    tokens = tuple([Token(token_type=token.token_type, text=token.text) for token in tokenize(sql, dialect=dialect)])
    return ParserState.from_tokens(tokens) 

def token(token_type: Optional[Enum] = None, 
          text: Optional[str] = None, 
          case_sensitive: bool = False,
          regex: Optional[re.Pattern[str]] = None
          ) -> Syntax[Any, Any]:
    token_type_txt = token_type.name if token_type is not None else None
    token_value_txt = text if text is not None else None
    msg = 'token(' + ','.join([x for x in [token_type_txt, token_value_txt, str(regex)] if x is not None]) + ')'
    return Syntax(
        lambda cls: cls.factory('token', token_type=token_type, text=text, case_sensitive=case_sensitive, regex=regex)
        ).describe(name=msg, fixity='prefix') 

    
def identifier(value: str | None = None) -> Syntax[Any, Any]:
    if value is None:
        return token(TokenType.IDENTIFIER)
    else:
        return token(TokenType.IDENTIFIER, text=value)

def variable(value: str | None = None) -> Syntax[Any, Any]:
    if value is None:
        return token(TokenType.VAR)
    else:
        return token(TokenType.VAR, text=value)

def literal(lit: str) -> Syntax[Any, Any]:
    return token(token_type=None, text=lit, case_sensitive=True)

def regex(regex: re.Pattern[str]) -> Syntax[Any, Any]:
    return token(token_type=None, regex=regex, case_sensitive=True)

def lift(value: Any)-> Syntax[Any, Any]:
    if isinstance(value, str):
        return literal(value)
    elif isinstance(value, re.Pattern):
        return token(regex=value)
    elif isinstance(value, Enum):
        return token(value)
    else:
        return Syntax(lambda cls: cls.success(value))

def number() -> Syntax[Any, Any]:
    return token(TokenType.NUMBER)


def string() -> Syntax[Any, Any]:
    return token(TokenType.STRING)



def until(*open_close: Tuple[Syntax[Tuple[T, ...] | T, ParserState[T]], Syntax[Tuple[T, ...] | T, ParserState[T]]],
          terminator: Optional[Syntax[Tuple[T, ...] | T, ParserState[T]]] = None,
          inclusive: bool = True, 
          strict: bool = True) -> Syntax[Any, Any]:
    return Syntax(
        lambda cls: cls.factory('until', 
                           *[(left.alg(cls), right.alg(cls)) for left, right in open_close], 
                           terminator=terminator.alg(cls) if terminator else None, 
                           inclusive=inclusive, 
                           strict=strict)
        ).describe(name="until", fixity='prefix') 






from __future__ import annotations
import re
from typing import (
    Optional, Any, TypeVar, Tuple, runtime_checkable, 
    Dict, Generic, Callable, Union, Protocol
)


from dataclasses import dataclass
from enum import Enum
from syncraft.constraint import Bindable



A = TypeVar('A')
B = TypeVar('B')  
C = TypeVar('C')  
S = TypeVar('S', bound=Bindable)  

@dataclass(frozen=True)
class Reducer(Generic[A, S]):
    run_f: Callable[[A, S], S]
    def __call__(self, a: A, s: S) -> S:
        return self.run_f(a, s)
    
    def map(self, f: Callable[[B], A]) -> Reducer[B, S]:
        def map_run(b: B, s: S) -> S:
            return self(f(b), s)
        return Reducer(map_run)
    

    
@dataclass(frozen=True)
class Bimap(Generic[A, B]):
    run_f: Callable[[A], Tuple[B, Callable[[B], A]]]
    def __call__(self, a: A) -> Tuple[B, Callable[[B], A]]:
        return self.run_f(a)    
    def __rshift__(self, other: Bimap[B, C]) -> Bimap[A, C]:
        def then_run(a: A) -> Tuple[C, Callable[[C], A]]:
            b, inv1 = self(a)
            c, inv2 = other(b)
            def inv(c2: C) -> A:
                return inv1(inv2(c2))
            return c, inv
        return Bimap(then_run)
    @staticmethod
    def const(a: B)->Bimap[B, B]:
        return Bimap(lambda _: (a, lambda b: b))

    @staticmethod
    def identity()->Bimap[Any, Any]:
        return Bimap(lambda a: (a, lambda b: b))

    @staticmethod
    def when(cond: Callable[[A], bool],
             then: Bimap[A, B],
             otherwise: Optional[Bimap[A, C]] = None) -> Bimap[A, A | B | C]:
        def when_run(a:A) -> Tuple[A | B | C, Callable[[A | B | C], A]]:
            bimap = then if cond(a) else (otherwise if otherwise is not None else Bimap.identity())
            abc, inv = bimap(a)
            def inv_f(b: Any) -> A:
                return inv(b)
            return abc, inv_f
        return Bimap(when_run)
    
    

@dataclass(frozen=True)
class Biarrow(Generic[S, A, B]):
    forward: Callable[[S, A], Tuple[S, B]]
    inverse: Callable[[S, B], Tuple[S, A]]
    def __rshift__(self, other: Biarrow[S, B, C]) -> Biarrow[S, A, C]:
        def fwd(s: S, a: A) -> Tuple[S, C]:
            s1, b = self.forward(s, a)
            return other.forward(s1, b)
        def inv(s: S, c: C) -> Tuple[S, A]:
            s1, b = other.inverse(s, c)
            return self.inverse(s1, b)
        return Biarrow(
            forward=fwd,
            inverse=inv
        )
    @staticmethod
    def identity()->Biarrow[S, A, A]:
        return Biarrow(
            forward=lambda s, x: (s, x),
            inverse=lambda s, y: (s, y)
        )
            
    @staticmethod
    def when(condition: Callable[..., bool], 
             then: Biarrow[S, A, B], 
             otherwise: Optional[Biarrow[S, A, B]] = None) -> Callable[..., Biarrow[S, A, B]]:
        def _when(*args:Any, **kwargs:Any) -> Biarrow[S, A, B]:
            return then if condition(*args, **kwargs) else (otherwise or Biarrow.identity())
        return _when
    

@dataclass(frozen=True)    
class AST:
    pass

class ChoiceKind(Enum):
    LEFT = 'left'
    RIGHT = 'right'
    
@dataclass(frozen=True)
class Choice(Generic[A, B], AST):
    kind: ChoiceKind
    left: Optional[A] 
    right: Optional[B] 


@dataclass(frozen=True)
class Many(Generic[A], AST):
    value: Tuple[A, ...]

@dataclass(frozen=True)
class Marked(Generic[A], AST):
    name: str
    value: A
class ThenKind(Enum):
    BOTH = '+'
    LEFT = '//'
    RIGHT = '>>'
    
FlatThen = Tuple[Any, ...]
MarkedThen = Tuple[Dict[str, Any] | Any, FlatThen]

@dataclass(eq=True, frozen=True)
class Then(Generic[A, B], AST):
    kind: ThenKind
    left: Optional[A]
    right: Optional[B]
        
@runtime_checkable
class TokenProtocol(Protocol):
    @property
    def token_type(self) -> Enum: ...
    @property
    def text(self) -> str: ...
    

@dataclass(frozen=True)
class Token:
    token_type: Enum
    text: str
    def __str__(self) -> str:
        return f"{self.token_type.name}({self.text})"
    
    def __repr__(self) -> str:
        return self.__str__()

    

@dataclass(frozen=True)
class TokenSpec:
    token_type: Optional[Enum] = None
    text: Optional[str] = None
    case_sensitive: bool = False
    regex: Optional[re.Pattern[str]] = None
        
    def is_valid(self, token: TokenProtocol) -> bool:
        type_match = self.token_type is None or token.token_type == self.token_type
        value_match = self.text is None or (token.text.strip() == self.text.strip() if self.case_sensitive else 
                                                    token.text.strip().upper() == self.text.strip().upper())
        value_match = value_match or (self.regex is not None and self.regex.fullmatch(token.text) is not None)
        return type_match and value_match




T = TypeVar('T', bound=TokenProtocol)  


ParseResult = Union[
    Then['ParseResult[T]', 'ParseResult[T]'], 
    Marked['ParseResult[T]'],
    Choice['ParseResult[T]', 'ParseResult[T]'],
    Many['ParseResult[T]'],
    T,
]






"""
A collection of parsers for SQL grammar using the syncraft library.
https://www.sqlite.org/syntaxdiagrams.html
"""
from __future__ import annotations
from typing import Any
from syncraft.syntax import Syntax, lazy, choice
import syncraft.parser as dsl
from syncraft.diagnostic import rich_error, rich_debug, rich_parser
from sqlglot import TokenType





L_PAREN = dsl.lift(TokenType.L_PAREN)
R_PAREN = dsl.lift(TokenType.R_PAREN)
L_BRACKET = dsl.lift(TokenType.L_BRACKET)
R_BRACKET = dsl.lift(TokenType.R_BRACKET)
L_BRACE = dsl.lift(TokenType.L_BRACE)
R_BRACE = dsl.lift(TokenType.R_BRACE)
COMMA = dsl.lift(TokenType.COMMA)
DOT = dsl.lift(TokenType.DOT)
DASH = dsl.lift(TokenType.DASH)
PLUS = dsl.lift(TokenType.PLUS)
COLON = dsl.lift(TokenType.COLON)
DOTCOLON = dsl.lift(TokenType.DOTCOLON)
DCOLON = dsl.lift(TokenType.DCOLON)
DQMARK = dsl.lift(TokenType.DQMARK)
SEMICOLON = dsl.lift(TokenType.SEMICOLON)
STAR = dsl.lift(TokenType.STAR)
BACKSLASH = dsl.lift(TokenType.BACKSLASH)
SLASH = dsl.lift(TokenType.SLASH)
LT = dsl.lift(TokenType.LT)
LTE = dsl.lift(TokenType.LTE)
GT = dsl.lift(TokenType.GT)
GTE = dsl.lift(TokenType.GTE)
NOT = dsl.lift(TokenType.NOT)
EQ = dsl.lift(TokenType.EQ)
NEQ = dsl.lift(TokenType.NEQ)
NULLSAFE_EQ = dsl.lift(TokenType.NULLSAFE_EQ)
COLON_EQ = dsl.lift(TokenType.COLON_EQ)
AND = dsl.lift(TokenType.AND)
OR = dsl.lift(TokenType.OR)
AMP = dsl.lift(TokenType.AMP)
DPIPE = dsl.lift(TokenType.DPIPE)
PIPE_GT = dsl.lift(TokenType.PIPE_GT)
PIPE = dsl.lift(TokenType.PIPE)
PIPE_SLASH = dsl.lift(TokenType.PIPE_SLASH)
DPIPE_SLASH = dsl.lift(TokenType.DPIPE_SLASH)
CARET = dsl.lift(TokenType.CARET)
CARET_AT = dsl.lift(TokenType.CARET_AT)
TILDA = dsl.lift(TokenType.TILDA)
ARROW = dsl.lift(TokenType.ARROW)
DARROW = dsl.lift(TokenType.DARROW)
FARROW = dsl.lift(TokenType.FARROW)
HASH = dsl.lift(TokenType.HASH)
HASH_ARROW = dsl.lift(TokenType.HASH_ARROW)
DHASH_ARROW = dsl.lift(TokenType.DHASH_ARROW)
LR_ARROW = dsl.lift(TokenType.LR_ARROW)
DAT = dsl.lift(TokenType.DAT)
LT_AT = dsl.lift(TokenType.LT_AT)
AT_GT = dsl.lift(TokenType.AT_GT)
DOLLAR = dsl.lift(TokenType.DOLLAR)
PARAMETER = dsl.lift(TokenType.PARAMETER)
SESSION_PARAMETER = dsl.lift(TokenType.SESSION_PARAMETER)
DAMP = dsl.lift(TokenType.DAMP)
XOR = dsl.lift(TokenType.XOR)
DSTAR = dsl.lift(TokenType.DSTAR)
URI_START = dsl.lift(TokenType.URI_START)
BLOCK_START = dsl.lift(TokenType.BLOCK_START)
BLOCK_END = dsl.lift(TokenType.BLOCK_END)
SPACE = dsl.lift(TokenType.SPACE)
BREAK = dsl.lift(TokenType.BREAK)
STRING = dsl.lift(TokenType.STRING)
NUMBER = dsl.lift(TokenType.NUMBER)
IDENTIFIER = dsl.lift(TokenType.IDENTIFIER)
DATABASE = dsl.lift(TokenType.DATABASE)
COLUMN = dsl.lift(TokenType.COLUMN)
COLUMN_DEF = dsl.lift(TokenType.COLUMN_DEF)
SCHEMA = dsl.lift(TokenType.SCHEMA)
TABLE = dsl.lift(TokenType.TABLE)
WAREHOUSE = dsl.lift(TokenType.WAREHOUSE)
STAGE = dsl.lift(TokenType.STAGE)
STREAMLIT = dsl.lift(TokenType.STREAMLIT)
VAR = dsl.lift(TokenType.VAR)
BIT_STRING = dsl.lift(TokenType.BIT_STRING)
HEX_STRING = dsl.lift(TokenType.HEX_STRING)
BYTE_STRING = dsl.lift(TokenType.BYTE_STRING)
NATIONAL_STRING = dsl.lift(TokenType.NATIONAL_STRING)
RAW_STRING = dsl.lift(TokenType.RAW_STRING)
HEREDOC_STRING = dsl.lift(TokenType.HEREDOC_STRING)
UNICODE_STRING = dsl.lift(TokenType.UNICODE_STRING)
BIT = dsl.lift(TokenType.BIT)
BOOLEAN = dsl.lift(TokenType.BOOLEAN)
TINYINT = dsl.lift(TokenType.TINYINT)
UTINYINT = dsl.lift(TokenType.UTINYINT)
SMALLINT = dsl.lift(TokenType.SMALLINT)
USMALLINT = dsl.lift(TokenType.USMALLINT)
MEDIUMINT = dsl.lift(TokenType.MEDIUMINT)
UMEDIUMINT = dsl.lift(TokenType.UMEDIUMINT)
INT = dsl.lift(TokenType.INT)
UINT = dsl.lift(TokenType.UINT)
BIGINT = dsl.lift(TokenType.BIGINT)
UBIGINT = dsl.lift(TokenType.UBIGINT)
INT128 = dsl.lift(TokenType.INT128)
UINT128 = dsl.lift(TokenType.UINT128)
INT256 = dsl.lift(TokenType.INT256)
UINT256 = dsl.lift(TokenType.UINT256)
FLOAT = dsl.lift(TokenType.FLOAT)
DOUBLE = dsl.lift(TokenType.DOUBLE)
UDOUBLE = dsl.lift(TokenType.UDOUBLE)
DECIMAL = dsl.lift(TokenType.DECIMAL)
DECIMAL32 = dsl.lift(TokenType.DECIMAL32)
DECIMAL64 = dsl.lift(TokenType.DECIMAL64)
DECIMAL128 = dsl.lift(TokenType.DECIMAL128)
DECIMAL256 = dsl.lift(TokenType.DECIMAL256)
UDECIMAL = dsl.lift(TokenType.UDECIMAL)
BIGDECIMAL = dsl.lift(TokenType.BIGDECIMAL)
CHAR = dsl.lift(TokenType.CHAR)
NCHAR = dsl.lift(TokenType.NCHAR)
VARCHAR = dsl.lift(TokenType.VARCHAR)
NVARCHAR = dsl.lift(TokenType.NVARCHAR)
BPCHAR = dsl.lift(TokenType.BPCHAR)
TEXT = dsl.lift(TokenType.TEXT)
MEDIUMTEXT = dsl.lift(TokenType.MEDIUMTEXT)
LONGTEXT = dsl.lift(TokenType.LONGTEXT)
BLOB = dsl.lift(TokenType.BLOB)
MEDIUMBLOB = dsl.lift(TokenType.MEDIUMBLOB)
LONGBLOB = dsl.lift(TokenType.LONGBLOB)
TINYBLOB = dsl.lift(TokenType.TINYBLOB)
TINYTEXT = dsl.lift(TokenType.TINYTEXT)
NAME = dsl.lift(TokenType.NAME)
BINARY = dsl.lift(TokenType.BINARY)
VARBINARY = dsl.lift(TokenType.VARBINARY)
JSON = dsl.lift(TokenType.JSON)
JSONB = dsl.lift(TokenType.JSONB)
TIME = dsl.lift(TokenType.TIME)
TIMETZ = dsl.lift(TokenType.TIMETZ)
TIMESTAMP = dsl.lift(TokenType.TIMESTAMP)
TIMESTAMPTZ = dsl.lift(TokenType.TIMESTAMPTZ)
TIMESTAMPLTZ = dsl.lift(TokenType.TIMESTAMPLTZ)
TIMESTAMPNTZ = dsl.lift(TokenType.TIMESTAMPNTZ)
TIMESTAMP_S = dsl.lift(TokenType.TIMESTAMP_S)
TIMESTAMP_MS = dsl.lift(TokenType.TIMESTAMP_MS)
TIMESTAMP_NS = dsl.lift(TokenType.TIMESTAMP_NS)
DATETIME = dsl.lift(TokenType.DATETIME)
DATETIME2 = dsl.lift(TokenType.DATETIME2)
DATETIME64 = dsl.lift(TokenType.DATETIME64)
SMALLDATETIME = dsl.lift(TokenType.SMALLDATETIME)
DATE = dsl.lift(TokenType.DATE)
DATE32 = dsl.lift(TokenType.DATE32)
INT4RANGE = dsl.lift(TokenType.INT4RANGE)
INT4MULTIRANGE = dsl.lift(TokenType.INT4MULTIRANGE)
INT8RANGE = dsl.lift(TokenType.INT8RANGE)
INT8MULTIRANGE = dsl.lift(TokenType.INT8MULTIRANGE)
NUMRANGE = dsl.lift(TokenType.NUMRANGE)
NUMMULTIRANGE = dsl.lift(TokenType.NUMMULTIRANGE)
TSRANGE = dsl.lift(TokenType.TSRANGE)
TSMULTIRANGE = dsl.lift(TokenType.TSMULTIRANGE)
TSTZRANGE = dsl.lift(TokenType.TSTZRANGE)
TSTZMULTIRANGE = dsl.lift(TokenType.TSTZMULTIRANGE)
DATERANGE = dsl.lift(TokenType.DATERANGE)
DATEMULTIRANGE = dsl.lift(TokenType.DATEMULTIRANGE)
UUID = dsl.lift(TokenType.UUID)
GEOGRAPHY = dsl.lift(TokenType.GEOGRAPHY)
NULLABLE = dsl.lift(TokenType.NULLABLE)
GEOMETRY = dsl.lift(TokenType.GEOMETRY)
POINT = dsl.lift(TokenType.POINT)
RING = dsl.lift(TokenType.RING)
LINESTRING = dsl.lift(TokenType.LINESTRING)
MULTILINESTRING = dsl.lift(TokenType.MULTILINESTRING)
POLYGON = dsl.lift(TokenType.POLYGON)
MULTIPOLYGON = dsl.lift(TokenType.MULTIPOLYGON)
HLLSKETCH = dsl.lift(TokenType.HLLSKETCH)
HSTORE = dsl.lift(TokenType.HSTORE)
SUPER = dsl.lift(TokenType.SUPER)
SERIAL = dsl.lift(TokenType.SERIAL)
SMALLSERIAL = dsl.lift(TokenType.SMALLSERIAL)
BIGSERIAL = dsl.lift(TokenType.BIGSERIAL)
XML = dsl.lift(TokenType.XML)
YEAR = dsl.lift(TokenType.YEAR)
USERDEFINED = dsl.lift(TokenType.USERDEFINED)
MONEY = dsl.lift(TokenType.MONEY)
SMALLMONEY = dsl.lift(TokenType.SMALLMONEY)
ROWVERSION = dsl.lift(TokenType.ROWVERSION)
IMAGE = dsl.lift(TokenType.IMAGE)
VARIANT = dsl.lift(TokenType.VARIANT)
OBJECT = dsl.lift(TokenType.OBJECT)
INET = dsl.lift(TokenType.INET)
IPADDRESS = dsl.lift(TokenType.IPADDRESS)
IPPREFIX = dsl.lift(TokenType.IPPREFIX)
IPV4 = dsl.lift(TokenType.IPV4)
IPV6 = dsl.lift(TokenType.IPV6)
ENUM = dsl.lift(TokenType.ENUM)
ENUM8 = dsl.lift(TokenType.ENUM8)
ENUM16 = dsl.lift(TokenType.ENUM16)
FIXEDSTRING = dsl.lift(TokenType.FIXEDSTRING)
LOWCARDINALITY = dsl.lift(TokenType.LOWCARDINALITY)
NESTED = dsl.lift(TokenType.NESTED)
AGGREGATEFUNCTION = dsl.lift(TokenType.AGGREGATEFUNCTION)
SIMPLEAGGREGATEFUNCTION = dsl.lift(TokenType.SIMPLEAGGREGATEFUNCTION)
TDIGEST = dsl.lift(TokenType.TDIGEST)
UNKNOWN = dsl.lift(TokenType.UNKNOWN)
VECTOR = dsl.lift(TokenType.VECTOR)
DYNAMIC = dsl.lift(TokenType.DYNAMIC)
VOID = dsl.lift(TokenType.VOID)
ALIAS = dsl.lift(TokenType.ALIAS)
ALTER = dsl.lift(TokenType.ALTER)
ALWAYS = dsl.lift(TokenType.ALWAYS)
ALL = dsl.lift(TokenType.ALL)
ANTI = dsl.lift(TokenType.ANTI)
ANY = dsl.lift(TokenType.ANY)
APPLY = dsl.lift(TokenType.APPLY)
ARRAY = dsl.lift(TokenType.ARRAY)
ASC = dsl.lift(TokenType.ASC)
ASOF = dsl.lift(TokenType.ASOF)
ATTACH = dsl.lift(TokenType.ATTACH)
AUTO_INCREMENT = dsl.lift(TokenType.AUTO_INCREMENT)
BEGIN = dsl.lift(TokenType.BEGIN)
BETWEEN = dsl.lift(TokenType.BETWEEN)
BULK_COLLECT_INTO = dsl.lift(TokenType.BULK_COLLECT_INTO)
CACHE = dsl.lift(TokenType.CACHE)
CASE = dsl.lift(TokenType.CASE)
CHARACTER_SET = dsl.lift(TokenType.CHARACTER_SET)
CLUSTER_BY = dsl.lift(TokenType.CLUSTER_BY)
COLLATE = dsl.lift(TokenType.COLLATE)
COMMAND = dsl.lift(TokenType.COMMAND)
COMMENT = dsl.lift(TokenType.COMMENT)
COMMIT = dsl.lift(TokenType.COMMIT)
CONNECT_BY = dsl.lift(TokenType.CONNECT_BY)
CONSTRAINT = dsl.lift(TokenType.CONSTRAINT)
COPY = dsl.lift(TokenType.COPY)
CREATE = dsl.lift(TokenType.CREATE)
CROSS = dsl.lift(TokenType.CROSS)
CUBE = dsl.lift(TokenType.CUBE)
CURRENT_DATE = dsl.lift(TokenType.CURRENT_DATE)
CURRENT_DATETIME = dsl.lift(TokenType.CURRENT_DATETIME)
CURRENT_SCHEMA = dsl.lift(TokenType.CURRENT_SCHEMA)
CURRENT_TIME = dsl.lift(TokenType.CURRENT_TIME)
CURRENT_TIMESTAMP = dsl.lift(TokenType.CURRENT_TIMESTAMP)
CURRENT_USER = dsl.lift(TokenType.CURRENT_USER)
DECLARE = dsl.lift(TokenType.DECLARE)
DEFAULT = dsl.lift(TokenType.DEFAULT)
DELETE = dsl.lift(TokenType.DELETE)
DESC = dsl.lift(TokenType.DESC)
DESCRIBE = dsl.lift(TokenType.DESCRIBE)
DETACH = dsl.lift(TokenType.DETACH)
DICTIONARY = dsl.lift(TokenType.DICTIONARY)
DISTINCT = dsl.lift(TokenType.DISTINCT)
DISTRIBUTE_BY = dsl.lift(TokenType.DISTRIBUTE_BY)
DIV = dsl.lift(TokenType.DIV)
DROP = dsl.lift(TokenType.DROP)
ELSE = dsl.lift(TokenType.ELSE)
END = dsl.lift(TokenType.END)
ESCAPE = dsl.lift(TokenType.ESCAPE)
EXCEPT = dsl.lift(TokenType.EXCEPT)
EXECUTE = dsl.lift(TokenType.EXECUTE)
EXISTS = dsl.lift(TokenType.EXISTS)
FALSE = dsl.lift(TokenType.FALSE)
FETCH = dsl.lift(TokenType.FETCH)
FILE_FORMAT = dsl.lift(TokenType.FILE_FORMAT)
FILTER = dsl.lift(TokenType.FILTER)
FINAL = dsl.lift(TokenType.FINAL)
FIRST = dsl.lift(TokenType.FIRST)
FOR = dsl.lift(TokenType.FOR)
FORCE = dsl.lift(TokenType.FORCE)
FOREIGN_KEY = dsl.lift(TokenType.FOREIGN_KEY)
FORMAT = dsl.lift(TokenType.FORMAT)
FROM = dsl.lift(TokenType.FROM)
FULL = dsl.lift(TokenType.FULL)
FUNCTION = dsl.lift(TokenType.FUNCTION)
GET = dsl.lift(TokenType.GET)
GLOB = dsl.lift(TokenType.GLOB)
GLOBAL = dsl.lift(TokenType.GLOBAL)
GRANT = dsl.lift(TokenType.GRANT)
GROUP_BY = dsl.lift(TokenType.GROUP_BY)
GROUPING_SETS = dsl.lift(TokenType.GROUPING_SETS)
HAVING = dsl.lift(TokenType.HAVING)
HINT = dsl.lift(TokenType.HINT)
IGNORE = dsl.lift(TokenType.IGNORE)
ILIKE = dsl.lift(TokenType.ILIKE)

IN = dsl.lift(TokenType.IN)
INDEX = dsl.lift(TokenType.INDEX)
INNER = dsl.lift(TokenType.INNER)
INSERT = dsl.lift(TokenType.INSERT)
INTERSECT = dsl.lift(TokenType.INTERSECT)
INTERVAL = dsl.lift(TokenType.INTERVAL)
INTO = dsl.lift(TokenType.INTO)
INTRODUCER = dsl.lift(TokenType.INTRODUCER)
IRLIKE = dsl.lift(TokenType.IRLIKE)
IS = dsl.lift(TokenType.IS)
ISNULL = dsl.lift(TokenType.ISNULL)
JOIN = dsl.lift(TokenType.JOIN)
JOIN_MARKER = dsl.lift(TokenType.JOIN_MARKER)
KEEP = dsl.lift(TokenType.KEEP)
KEY = dsl.lift(TokenType.KEY)
KILL = dsl.lift(TokenType.KILL)
LANGUAGE = dsl.lift(TokenType.LANGUAGE)
LATERAL = dsl.lift(TokenType.LATERAL)
LEFT = dsl.lift(TokenType.LEFT)
LIKE = dsl.lift(TokenType.LIKE)

LIMIT = dsl.lift(TokenType.LIMIT)
LIST = dsl.lift(TokenType.LIST)
LOAD = dsl.lift(TokenType.LOAD)
LOCK = dsl.lift(TokenType.LOCK)
MAP = dsl.lift(TokenType.MAP)
MATCH_CONDITION = dsl.lift(TokenType.MATCH_CONDITION)
MATCH_RECOGNIZE = dsl.lift(TokenType.MATCH_RECOGNIZE)
MEMBER_OF = dsl.lift(TokenType.MEMBER_OF)
MERGE = dsl.lift(TokenType.MERGE)
MOD = dsl.lift(TokenType.MOD)
MODEL = dsl.lift(TokenType.MODEL)
NATURAL = dsl.lift(TokenType.NATURAL)
NEXT = dsl.lift(TokenType.NEXT)
NOTHING = dsl.lift(TokenType.NOTHING)
NOTNULL = dsl.lift(TokenType.NOTNULL)
NULL = dsl.lift(TokenType.NULL)
OBJECT_IDENTIFIER = dsl.lift(TokenType.OBJECT_IDENTIFIER)
OFFSET = dsl.lift(TokenType.OFFSET)
ON = dsl.lift(TokenType.ON)
ONLY = dsl.lift(TokenType.ONLY)
OPERATOR = dsl.lift(TokenType.OPERATOR)
ORDER_BY = dsl.lift(TokenType.ORDER_BY)
ORDER_SIBLINGS_BY = dsl.lift(TokenType.ORDER_SIBLINGS_BY)
ORDERED = dsl.lift(TokenType.ORDERED)
ORDINALITY = dsl.lift(TokenType.ORDINALITY)
OUTER = dsl.lift(TokenType.OUTER)
OVER = dsl.lift(TokenType.OVER)
OVERLAPS = dsl.lift(TokenType.OVERLAPS)
OVERWRITE = dsl.lift(TokenType.OVERWRITE)
PARTITION = dsl.lift(TokenType.PARTITION)
PARTITION_BY = dsl.lift(TokenType.PARTITION_BY)
PERCENT = dsl.lift(TokenType.PERCENT)
PIVOT = dsl.lift(TokenType.PIVOT)
PLACEHOLDER = dsl.lift(TokenType.PLACEHOLDER)
POSITIONAL = dsl.lift(TokenType.POSITIONAL)
PRAGMA = dsl.lift(TokenType.PRAGMA)
PREWHERE = dsl.lift(TokenType.PREWHERE)
PRIMARY_KEY = dsl.lift(TokenType.PRIMARY_KEY)
PROCEDURE = dsl.lift(TokenType.PROCEDURE)
PROPERTIES = dsl.lift(TokenType.PROPERTIES)
PSEUDO_TYPE = dsl.lift(TokenType.PSEUDO_TYPE)
PUT = dsl.lift(TokenType.PUT)
QUALIFY = dsl.lift(TokenType.QUALIFY)
QUOTE = dsl.lift(TokenType.QUOTE)
RANGE = dsl.lift(TokenType.RANGE)
RECURSIVE = dsl.lift(TokenType.RECURSIVE)
REFRESH = dsl.lift(TokenType.REFRESH)
RENAME = dsl.lift(TokenType.RENAME)
REPLACE = dsl.lift(TokenType.REPLACE)
RETURNING = dsl.lift(TokenType.RETURNING)
REFERENCES = dsl.lift(TokenType.REFERENCES)
RIGHT = dsl.lift(TokenType.RIGHT)
RLIKE = dsl.lift(TokenType.RLIKE)
ROLLBACK = dsl.lift(TokenType.ROLLBACK)
ROLLUP = dsl.lift(TokenType.ROLLUP)
ROW = dsl.lift(TokenType.ROW)
ROWS = dsl.lift(TokenType.ROWS)
SELECT = dsl.lift(TokenType.SELECT)
SEMI = dsl.lift(TokenType.SEMI)
SEPARATOR = dsl.lift(TokenType.SEPARATOR)
SEQUENCE = dsl.lift(TokenType.SEQUENCE)
SERDE_PROPERTIES = dsl.lift(TokenType.SERDE_PROPERTIES)
SET = dsl.lift(TokenType.SET)
SETTINGS = dsl.lift(TokenType.SETTINGS)
SHOW = dsl.lift(TokenType.SHOW)
SIMILAR_TO = dsl.lift(TokenType.SIMILAR_TO)
SOME = dsl.lift(TokenType.SOME)
SORT_BY = dsl.lift(TokenType.SORT_BY)
START_WITH = dsl.lift(TokenType.START_WITH)
STORAGE_INTEGRATION = dsl.lift(TokenType.STORAGE_INTEGRATION)
STRAIGHT_JOIN = dsl.lift(TokenType.STRAIGHT_JOIN)
STRUCT = dsl.lift(TokenType.STRUCT)
SUMMARIZE = dsl.lift(TokenType.SUMMARIZE)
TABLE_SAMPLE = dsl.lift(TokenType.TABLE_SAMPLE)
TAG = dsl.lift(TokenType.TAG)
TEMPORARY = dsl.lift(TokenType.TEMPORARY)    
TOP = dsl.lift(TokenType.TOP)
THEN = dsl.lift(TokenType.THEN)
TRUE = dsl.lift(TokenType.TRUE)
TRUNCATE = dsl.lift(TokenType.TRUNCATE)
UNCACHE = dsl.lift(TokenType.UNCACHE)
UNION = dsl.lift(TokenType.UNION)
UNNEST = dsl.lift(TokenType.UNNEST)
UNPIVOT = dsl.lift(TokenType.UNPIVOT)
UPDATE = dsl.lift(TokenType.UPDATE)
USE = dsl.lift(TokenType.USE)
USING = dsl.lift(TokenType.USING)
VALUES = dsl.lift(TokenType.VALUES)
VIEW = dsl.lift(TokenType.VIEW)
VOLATILE = dsl.lift(TokenType.VOLATILE)
WHEN = dsl.lift(TokenType.WHEN)
WHERE = dsl.lift(TokenType.WHERE)
WINDOW = dsl.lift(TokenType.WINDOW)
WITH = dsl.lift(TokenType.WITH)
UNIQUE = dsl.lift(TokenType.UNIQUE)
VERSION_SNAPSHOT = dsl.lift(TokenType.VERSION_SNAPSHOT)
TIMESTAMP_SNAPSHOT = dsl.lift(TokenType.TIMESTAMP_SNAPSHOT)
OPTION = dsl.lift(TokenType.OPTION)
SINK = dsl.lift(TokenType.SINK)
SOURCE = dsl.lift(TokenType.SOURCE)
ANALYZE = dsl.lift(TokenType.ANALYZE)
NAMESPACE = dsl.lift(TokenType.NAMESPACE)
EXPORT = dsl.lift(TokenType.EXPORT)
HIVE_TOKEN_STREAM = dsl.lift(TokenType.HIVE_TOKEN_STREAM)

ABORT = dsl.lift("ABORT")
FAIL = dsl.lift("FAIL")
LOOP = dsl.lift("LOOP")
WHILE = dsl.lift("WHILE")
TRIGGER  = dsl.lift("TRIGGER")
TEMP = dsl.lift("TEMP")
IF = dsl.lift("IF")

BEFORE = dsl.lift("BEFORE")
AFTER = dsl.lift("AFTER")
INSTEAD = dsl.lift("INSTEAD")
OF = dsl.lift("OF")
EACH = dsl.lift("EACH")

ADD = dsl.lift("ADD")
TO = dsl.lift("TO")

RAISE = dsl.lift("RAISE")
RETURNS = dsl.lift("RETURNS")
PRIMARY = dsl.lift("PRIMARY")
NULLS = dsl.lift("NULLS")
LAST = dsl.lift("LAST")
CONFLICT = dsl.lift("CONFLICT")
CHECK = dsl.lift("CHECK")
GENERATED = dsl.lift("GENERATED")
STORED = dsl.lift("STORED")
VIRTUAL = dsl.lift("VIRTUAL")
AS = ALIAS
CASCADE = dsl.lift("CASCADE")
RESTRICT = dsl.lift("RESTRICT")
NO = dsl.lift("NO")
ACTION = dsl.lift("ACTION")
NO_ACTION = NO >> ACTION
MATCH = dsl.lift("MATCH")
DEFERRABLE = dsl.lift("DEFERRABLE")
INITIALLY = dsl.lift("INITIALLY")
IMMEDIATE = dsl.lift("IMMEDIATE")
DEFERRED = dsl.lift("DEFERRED")
RELY = dsl.lift("RELY")
NORELY = dsl.lift("NORELY")
VALIDATE = dsl.lift("VALIDATE")
NOVALIDATE = dsl.lift("NOVALIDATE")
EXCLUSIVE = dsl.lift("EXCLUSIVE")
TRANSACTION = dsl.lift("TRANSACTION")
WITHOUT = dsl.lift("WITHOUT")
ROWID = dsl.lift("ROWID")
STRICT = dsl.lift("STRICT")
MATERIALIZED = dsl.lift("MATERIALIZED")
DO = dsl.lift("DO")
RELEASE = dsl.lift("RELEASE")
SAVEPOINT = dsl.lift("SAVEPOINT")
REINDEX = dsl.lift("REINDEX")
INDEXED = dsl.lift("INDEXED")
VACUUM = dsl.lift("VACUUM")
GROUP = dsl.lift("GROUP")
GROUPS = dsl.lift("GROUPS")
UNBOUNDED = dsl.lift("UNBOUNDED")
PRECEDING = dsl.lift("PRECEDING")
FOLLOWING = dsl.lift("FOLLOWING")
CURRENT = dsl.lift("CURRENT")
EXCLUDE = dsl.lift("EXCLUDE")
OTHERS = dsl.lift("OTHERS")
TIES = dsl.lift("TIES")
BY = dsl.lift("BY")
CAST = dsl.lift("CAST")
REGEXP = dsl.lift("REGEXP")

var = dsl.variable()
string = dsl.string()
number = dsl.number()

signed_number = ~(PLUS | DASH) + number
literal_value = (number | string | BLOB | NULL | TRUE | FALSE | CURRENT_DATE | CURRENT_TIME | CURRENT_TIMESTAMP)
if_not_exists = (IF >> NOT >> EXISTS)
if_exists = (IF >> EXISTS)
bind_parameter = ((PLACEHOLDER >> ~number) | (COLON >> var) | (PARAMETER >> var) | var)
schema_name = (var // DOT)
table_name = var
view_name = var
trigger_name = var
constraint_name = var
table_as_alias = (table_name // ~(~AS + var))
column_name = var
index_name = var
table_function_name = var
table_alias = var
alias = var
window_name = var
for_each_row = (FOR >> EACH >> ROW)
unary_operator = (PLUS | DASH)
binary_operator = (PLUS | DASH | STAR | SLASH | EQ | NEQ | GT | GTE | LT | LTE)
compound_operator = ((UNION >> ~ALL) | EXCEPT | INTERSECT)

collate_name = var
function_name = var
expr = lazy(lambda: expression())
frame_spec = ((RANGE | ROWS | GROUPS) >> (
                                                (UNBOUNDED >> PRECEDING)
                                                | (expr >> PRECEDING)
                                                | (CURRENT >> ROW)
                                                | (BETWEEN >> (
                                                    UNBOUNDED >> PRECEDING
                                                    | expr >> PRECEDING
                                                    | CURRENT >> ROW
                                                    | expr >> FOLLOWING
                                                ) >> AND >> (
                                                    expr >> PRECEDING
                                                    | CURRENT >> ROW
                                                    | expr >> FOLLOWING
                                                    | UNBOUNDED >> FOLLOWING
                                                ))
                                            ) >> ~(
                                                EXCLUDE >> ((CURRENT >> ROW) | GROUP | (NO >> OTHERS) | TIES)
                                            ))

join_operator = ((COMMA 
                    | JOIN 
                    | CROSS >> JOIN
                    | NATURAL >> (JOIN
                                    | INNER >> ~OUTER >>JOIN
                                    | LEFT >> ~OUTER >> JOIN
                                    | RIGHT >> ~OUTER >> JOIN
                                    | FULL >> ~OUTER >> JOIN
                                    )))

join_constraint = ~((ON >> expr) | (USING >> var.parens(COMMA, L_PAREN, R_PAREN)))

ordering_term = (expr >> ~(COLLATE >> collate_name) >> ~(ASC | DESC) >> ~(NULLS >> (LAST | FIRST)))
function_argument = (~STAR | (~DISTINCT >> expr.sep_by(COMMA) >> ~(ORDER_BY >> ordering_term.sep_by(COMMA))))
filter_clause = (FILTER >> (WHERE >> expr).between(L_PAREN, R_PAREN))
over_clause = (OVER >> ~(window_name | L_PAREN >> ~var >> (
                                                                            ~(PARTITION >> BY >> expr.sep_by(COMMA))
                                                                            >> ~(ORDER_BY >> ordering_term.sep_by(COMMA))
                                                                            >> ~frame_spec
                                                                        ) // R_PAREN))
                                                                            
typed_name = var >> ~ signed_number.parens(COMMA, L_PAREN, R_PAREN)
returning_clause = RETURNING >> (expr | STAR | (expr >> ~AS >> var)).sep_by(COMMA) 

select_stmt = lazy(lambda: select_statement())

common_table_expression = (table_name >> ~column_name.parens(COMMA, L_PAREN, R_PAREN) >> AS >> ~NOT >> ~MATERIALIZED >> select_stmt.between(L_PAREN, R_PAREN))

indexed_column = (expr | var) >> ~(COLLATE >> var) >> ~(ASC | DESC) 
upsert_clause = (ON >> CONFLICT >> ~(indexed_column.parens(COMMA, L_PAREN, R_PAREN) >> ~(WHERE >> expr))
            >> DO
            >> (
                NOTHING 
                | (UPDATE 
                   >> SET 
                   >> ((column_name 
                        | column_name.parens(COMMA, L_PAREN, R_PAREN)) 
                        >> EQ 
                        >> expr).sep_by(COMMA) 
                >> ~(WHERE >> expr))
            )
            ).many()

window_defn = L_PAREN >> ~window_name >> ~(PARTITION >> BY >> expr.sep_by(COMMA)) >> ~(ORDER_BY >> ordering_term.sep_by(COMMA) >> ~frame_spec) // R_PAREN
result_columns = ((expr >> ~(~AS >> var)) | STAR | (table_name >> DOT>>var))
table_subquery = lazy(lambda: table_or_subquery())
join_clause = (table_subquery >> ~((join_operator >> table_subquery >> join_constraint).many()))
indexed_column = (expr | column_name) >> ~(COLLATE >> collate_name) >> ~(ASC | DESC)
conflict_clause = ON >> CONFLICT >> (ROLLBACK | ABORT | FAIL | IGNORE | REPLACE)
foreign_key_clause = (REFERENCES 
                      >> table_name 
                      >> ~column_name.parens(COMMA, L_PAREN, R_PAREN) 
                      >> ((ON >> (DELETE | UPDATE) >> (
    (SET >> (NULL | DEFAULT)) | CASCADE | RESTRICT | NO_ACTION
)) | (MATCH  >> var)).many() >> ~(~NOT >> DEFERRABLE) >> ~(INITIALLY >> (DEFERRED | IMMEDIATE)))

qualified_table_name = ~schema_name >> table_name >> ~(AS >> alias) >> ~((INDEXED >> BY >> index_name) | (NOT >> INDEXED))

update_stmt = (
        WITH >> ~(RECURSIVE >> common_table_expression.sep_by(COMMA))>>
        UPDATE>>
        ~(OR >> (ABORT | IGNORE | FAIL | REPLACE | ROLLBACK))>>
        qualified_table_name>>
        SET >> (var | var.parens(COMMA, L_PAREN, R_PAREN)) >> EQ >> expr>>
        ~(FROM >> (table_subquery.sep_by(COMMA) | join_clause))>>
        ~(WHERE >> expr)>>
        ~returning_clause>>
        ~SEMICOLON
    )

update_stmt_limited = (
        WITH >> ~(RECURSIVE >> common_table_expression.sep_by(COMMA))>>
        UPDATE>>
        ~(OR >> (ABORT | IGNORE | FAIL | REPLACE | ROLLBACK))>>
        qualified_table_name>>
        SET >> ((column_name | column_name.parens(COMMA, L_PAREN, R_PAREN)) >> EQ >> expr).sep_by(COMMA)>>
        ~(FROM >> (table_subquery.sep_by(COMMA) | join_clause))>>
        ~(WHERE >> expr)>>
        ~returning_clause>>
        ~(ORDER_BY >> ordering_term.sep_by(COMMA))>>
        ~(LIMIT >> expr >> ~((OFFSET >> expr) | (COMMA >> expr)))>>
        ~SEMICOLON
    )


def table_or_subquery()->Syntax[Any, Any]:
    t1 = ~schema_name >> table_as_alias >> ~((INDEXED >> BY >> index_name)|(NOT >> INDEXED))
    t2 = ~schema_name >> table_function_name >> expr.parens(COMMA, L_PAREN, R_PAREN) >> ~(~AS >> var)
    t3 = select_stmt.between(L_PAREN, R_PAREN) >> ~(~AS >> var)
    t4 = table_subquery.parens(COMMA, L_PAREN, R_PAREN)
    t5 = join_clause.between(L_PAREN, R_PAREN) 
    return (t1 | t2 | t3 | t4 | t5).as_(Syntax[Any, Any])


def expression() -> Syntax[Any, Any]:
    return choice(
        literal_value,
        bind_parameter,
        ~(~schema_name >> table_name >> DOT) >> column_name, 
        unary_operator >> expr,
        expr >> binary_operator >> expr,
        function_name 
            >> function_argument.between(L_PAREN, R_PAREN) 
            >> ~filter_clause 
            >> ~over_clause,
        L_PAREN >> expr.sep_by(COMMA) // R_PAREN,
        CAST >> L_PAREN >> expr >> AS >> typed_name >> R_PAREN,
        expr >> COLLATE >> var,
        expr >> ~NOT >> LIKE >> expr >> ~(ESCAPE >> expr),
        expr >> ~NOT >> (GLOB | REGEXP | MATCH) >> expr,
        expr >> (ISNULL | NOTNULL | (NOT >> NULL)),
        expr >> IS >> ~NOT >> ~(DISTINCT >> FROM) >> expr,
        expr >> ~NOT >> BETWEEN >> expr >> (AND >> expr),
        expr >> ~NOT >> IN >> L_PAREN >> (expr.sep_by(COMMA) | select_stmt) // R_PAREN,
        expr >> ~NOT >> IN >> ~schema_name >> (table_name | (function_name >> expr.parens(COMMA, L_PAREN, R_PAREN))),
        ~NOT >> ~EXISTS >> select_stmt.between(L_PAREN, R_PAREN),
        CASE >> ~expr >> (WHEN >> expr >> THEN >> expr).many() >> ~(ELSE >> expr) // END,
    ).as_(Syntax[Any, Any])

def select_statement() -> Syntax[Any, Any]:
    select_clause = SELECT >> ~(DISTINCT | ALL) >> result_columns.sep_by(COMMA)
    from_clause = FROM >> (table_subquery.sep_by(COMMA) | join_clause)
    where_clause = WHERE >> expr
    having_clause = HAVING >> expr
    group_by_clause = GROUP >> BY >> expr.sep_by(COMMA)
    window_clause = WINDOW >> (window_name >> AS >> window_defn).sep_by(COMMA)
    value_clause = VALUES >> expr.parens(COMMA, L_PAREN, R_PAREN).sep_by(COMMA)
    limit_clause = LIMIT >> expr >> ~((OFFSET >> expr) | (COMMA >> expr))
    ordering_clause = ORDER_BY >> ordering_term.sep_by(COMMA)
    select_core = value_clause | (select_clause >> ~from_clause >> ~(where_clause >> ~having_clause) >> ~(group_by_clause >> ~having_clause) >> ~window_clause)
    return (
        WITH >> ~(RECURSIVE >> common_table_expression.sep_by(COMMA))
         >> select_core.sep_by(compound_operator)
         >> ~(ordering_clause >> ~limit_clause)
         >> ~SEMICOLON
    ).as_(Syntax[Any, Any])

column_constraint = ~(CONSTRAINT >> constraint_name) >> (
    (PRIMARY >> KEY >> ~(ASC | DESC) >> ~conflict_clause >> AUTO_INCREMENT)
    | (NOT >> NULL >> conflict_clause)
    | (UNIQUE >> conflict_clause)
    | (CHECK >> expr)
    | (DEFAULT >> (literal_value | signed_number | expr))
    | (COLLATE >> collate_name)
    | ~(GENERATED >> ALWAYS) >> AS >> expr // ~(STORED | VIRTUAL)
)

column_def = (
            var>>
            typed_name>>
            ~column_constraint>>
            ~ SEMICOLON
        )

table_options = ((WITHOUT >> ROWID) | STRICT).sep_by(COMMA)

table_constraint = ~(CONSTRAINT >> constraint_name) >> choice(
    (PRIMARY >> KEY >> ~(ASC | DESC) >> ~conflict_clause >> AUTO_INCREMENT),
    (UNIQUE >> conflict_clause),
    (CHECK >> expr),
    foreign_key_clause
)


rename_table_stmt = (
            ALTER>>
            TABLE>>
            ~schema_name>>
            table_name>>
            RENAME >> TO >> var>>
            ~ SEMICOLON
        )
    

rename_column_stmt = (
             ALTER>>
             TABLE>>
            ~schema_name>>
            table_name>>
             RENAME>>
             COLUMN>>
             column_name>>
             TO>>
            column_name>>
            ~ SEMICOLON
        )

add_column_stmt = (
            ALTER>>
            TABLE>>
            ~schema_name>>
            table_name>>
            ADD>>
            COLUMN>>
            column_name>>
            column_def>>
            ~ SEMICOLON
        )


drop_column_stmt = (
            ALTER>>
            TABLE>>
            ~schema_name>>
            table_name>>
            DROP>>
            COLUMN>>
            column_name>>
            ~ SEMICOLON
        )

alter_table_stmt = rename_table_stmt | rename_column_stmt | add_column_stmt | drop_column_stmt


analyze_stmt = (
            ANALYZE>>
            ~schema_name>>
            table_name>>
            ~SEMICOLON
        )
    

attach_stmt = (
            ATTACH>>
            ~ DATABASE>>
            expr>>
            ~(AS >> var)>>
            ~SEMICOLON
        )

begin_stmt = (
            BEGIN>>
            ~IMMEDIATE>>
            ~DEFERRED>>
            ~EXCLUSIVE>>
            ~TRANSACTION>>
            ~SEMICOLON
        )
    


commit_stmt = (
            (COMMIT | END)>>
            ~TRANSACTION>>
            ~SEMICOLON
        )
    

create_index_stmt = (
            CREATE>>
            ~(TEMPORARY | TEMP)>>
            ~UNIQUE>>
            INDEX>>
            ~if_not_exists>>
            ~schema_name>>
            index_name>>
            ON>>
            table_name>>
            column_name.parens(COMMA, L_PAREN, R_PAREN)>>
            ~(WHERE >> expr)>>
            ~SEMICOLON
        )
    

create_table_stmt = (
            CREATE>>
            ~(TEMPORARY | TEMP)>>
            TABLE>>
            ~if_not_exists>>
            ~schema_name>>
            table_name>>
            (L_PAREN >>  column_def.sep_by(COMMA) + ~(COMMA >> table_constraint.sep_by(COMMA)) // R_PAREN) | (AS >> select_stmt)>>
            ~table_options>>
            ~SEMICOLON
        )
    

create_view_stmt = (   
            CREATE>>
            ~(TEMPORARY | TEMP)>>
            VIEW>>
            ~if_not_exists>>
            ~schema_name>>
            table_name>>
            ~(L_PAREN >> var.sep_by(COMMA) // R_PAREN)>>
            AS>>
            select_stmt>>
            ~SEMICOLON
        )
    

create_virtual_table_stmt = (
            CREATE>>
            VIRTUAL>>
            TABLE>>
            ~if_not_exists>>
            ~schema_name>>
            table_name>>
            USING>>
            var>>
            ~(L_PAREN >> var.sep_by(COMMA) // R_PAREN)>>
            ~SEMICOLON
        )

delete_stmt = (
            WITH >> ~(RECURSIVE >> common_table_expression.sep_by(COMMA))>>
            DELETE>>
            ~FROM>>
            qualified_table_name>>
            ~(WHERE >> expr)>>
            ~returning_clause>>
            ~SEMICOLON
        )
    

delete_stmt_limited = (
            WITH >> ~(RECURSIVE >> common_table_expression.sep_by(COMMA))>>
            DELETE>>
            ~FROM>>
            qualified_table_name>>
            ~(WHERE >> expr)>>
            ~returning_clause>>
            ORDER_BY >> ordering_term.sep_by(COMMA)>>
            LIMIT >> expr >> ~((OFFSET >> expr) | (COMMA >> expr))>>
            ~SEMICOLON
        )

detach_stmt = (
            DETACH>>
            ~DATABASE>>
            schema_name>>
            ~SEMICOLON
        )


drop_index_stmt = (
            DROP>>
            INDEX>>
            ~if_exists>>
            ~schema_name>>
            index_name>>
            ~SEMICOLON
        )
    

drop_table_stmt = (
            DROP>>
            TABLE>>
            ~if_exists>>
            ~schema_name>>
            table_name>>
            ~SEMICOLON
        )
    

drop_view_stmt = (
            DROP>>
            VIEW>>
            ~if_exists>>
            ~schema_name>>
            view_name>>
            ~SEMICOLON
        )
    

drop_trigger_stmt = (
            DROP>>
            TRIGGER>>
            ~if_exists>>
            ~schema_name>>
            trigger_name>>
            ~SEMICOLON
        )
    

insert_stmt =(
            WITH >> ~(RECURSIVE >> common_table_expression.sep_by(COMMA))>>
            REPLACE | (INSERT >> ~(OR >> (ABORT | IGNORE | FAIL | REPLACE | ROLLBACK)))>>
            INTO>>
            ~schema_name>>
            table_name>>
            ~(AS >> var)>>
            ~(L_PAREN >> var.sep_by(COMMA) // R_PAREN)>>
            ~(VALUES >> expr.parens(COMMA, L_PAREN, R_PAREN).sep_by(COMMA) >> ~upsert_clause)>>
            ~(select_stmt >> ~upsert_clause)>>
            ~(DEFAULT >> VALUES)>>
            ~returning_clause>>
            ~SEMICOLON
        )
    

pragma_stmt = (
            PRAGMA>>
            ~schema_name>>
            var>>
            ~EQ>>
            ~((EQ >> (var | signed_number | literal_value)) | (var | signed_number | literal_value).between(L_PAREN, R_PAREN))>>
            ~SEMICOLON
        )

reindex_stmt = (
            REINDEX>>
            ~schema_name>>
            index_name>>
            ~SEMICOLON
        )
    

release_stmt = (
            RELEASE>>
            SAVEPOINT>>
            var>>
            ~SEMICOLON
        )
    

rollback_stmt = (
            ROLLBACK>>
            ~TRANSACTION>>
            ~TO>>
            ~SAVEPOINT>>
            ~(var // SEMICOLON)
        )

savepoint_stmt =(
            SAVEPOINT>>
            var>>
            ~SEMICOLON
        )
    



vacuum_stmt = (
            VACUUM>>
            ~var>>
            ~(INTO >> var)>>
            ~SEMICOLON
        )


create_trigger_stmt = (
            CREATE>>
            ~(TEMPORARY | TEMP)>>
            TRIGGER>>
            ~if_not_exists>>
            ~schema_name>>
            trigger_name>>
            BEFORE | AFTER | (INSTEAD >> OF)>>
            INSERT | DELETE | (UPDATE >> ~(OF >> var.sep_by(COMMA)))>>
            ON>>
            table_name>>
            ~for_each_row>>
            ~(WHEN >> expr)>>
            BEGIN>>
            ((update_stmt | insert_stmt | delete_stmt | select_stmt)>>SEMICOLON).many()>>
            END>>
            ~SEMICOLON
        )

raise_function = RAISE >> (IGNORE | ((ROLLBACK | FAIL | ABORT) >> COMMA >> expr))

aggregate_function_invocation = function_name >> (~STAR 
                                                  | expr 
                                                  | (~DISTINCT >> expr.sep_by(COMMA) >> ~(ORDER_BY >> ordering_term.sep_by(COMMA)))
                                                  ).between(L_PAREN, R_PAREN) >> ~filter_clause


sql_stmt = choice(
    alter_table_stmt,
    analyze_stmt,
    attach_stmt,
    begin_stmt,

    commit_stmt,
    create_index_stmt,    
    create_table_stmt,
    create_trigger_stmt,
    create_view_stmt,
    create_virtual_table_stmt,
    
    delete_stmt,
    delete_stmt_limited,
    detach_stmt,
    drop_index_stmt,
    drop_table_stmt,
    drop_trigger_stmt,
    drop_view_stmt,
    
    insert_stmt,
    pragma_stmt,
    reindex_stmt,
    release_stmt,
    rollback_stmt,
    savepoint_stmt,
    select_stmt,
    update_stmt,
    update_stmt_limited,
    vacuum_stmt,
)
    
if __name__ == "__main__":
    print("__all__ =", [str(token.name) for token in TokenType])
    for token in TokenType:
        print(f"{token.name} = lift({token})")




from __future__ import annotations
from typing import Callable, Generic, Tuple, TypeVar, Optional, Any, Self
from enum import Enum
from dataclasses import dataclass, field, replace
import collections.abc
from collections import defaultdict
from itertools import product

K = TypeVar('K')
V = TypeVar('V')
class FrozenDict(collections.abc.Mapping, Generic[K, V]):
    def __init__(self, *args, **kwargs):
        self._data = dict(*args, **kwargs)
        self._hash = None
    def __getitem__(self, key):
        return self._data[key]

    def __iter__(self):
        return iter(self._data)

    def __len__(self):
        return len(self._data)
        
    def __hash__(self):
        if self._hash is None:
            self._hash = hash(frozenset(self._data.items()))
        return self._hash

    def __eq__(self, other):
        if isinstance(other, collections.abc.Mapping):
            return self._data == other
        return NotImplemented

    def __repr__(self):
        return f"{self.__class__.__name__}({self._data})"




@dataclass(frozen=True)
class Expr:
    left: Any
    op: str
    right: Any


@dataclass(frozen=True)
class Variable:
    name: Optional[str] = None
    _root: Optional[Variable] = field(default=None, compare=False, repr=False)
    _mapf: Optional[Callable[[Any], Any]] = field(default=None, compare=False, repr=False)

    def __post_init__(self):
        if self._root is None:
            object.__setattr__(self, '_root', self)

    def raw(self, b:'BoundVar') -> Tuple[Any, ...]:
        assert self._root is not None, "_rawf can not be None"
        return b.get(self._root, ())
    

    def map(self, f: Callable[[Any], Any]) -> "Variable":
        if self._mapf is None:
            return replace(self, _mapf=f)
        else:
            oldf = self._mapf
            return replace(self, _mapf=lambda a: f(oldf(a)))
    
    def get(self, b: 'BoundVar') -> Tuple[Any, ...]:
        vals = self.raw(b)
        if self._mapf is not None:
            return tuple(self._mapf(v) for v in vals)
        else:
            return vals
        
    def __call__(self, b:'BoundVar', raw:bool=False) -> Any:
        if raw:
            return self.raw(b)
        else:
            return self.get(b)
        
    def __eq__(self, other): 
        return Expr(self, '==', other)
    def __ne__(self, other): 
        return Expr(self, '!=', other)
    def __lt__(self, other): 
        return Expr(self, '<', other)
    def __le__(self, other): 
        return Expr(self, '<=', other)
    def __gt__(self, other): 
        return Expr(self, '>', other)
    def __ge__(self, other): 
        return Expr(self, '>=', other)    

BoundVar = FrozenDict[Variable, Tuple[Any, ...]]


@dataclass(frozen=True)
class Binding:
    bindings : frozenset[Tuple[Variable, Any]] = frozenset()
    def bind(self, var: Variable, node: Any) -> Binding:
        new_binding = set(self.bindings)
        new_binding.add((var, node))
        return Binding(bindings=frozenset(new_binding))
    
    def to_dict(self)->BoundVar:
        ret = defaultdict(list)
        for var, node in self.bindings:
            ret[var].append(node)
        return FrozenDict({k: tuple(vs) for k, vs in ret.items()})


A = TypeVar('A')
@dataclass(frozen=True)
class Bindable:
    binding: Binding = field(default_factory=Binding)

    def map(self, f: Callable[[Any], Any])->Self: 
        return self
    
    def bind(self, var: Variable, node:Any)->Self:
        return replace(self, binding=self.binding.bind(var, node))


class Quantifier(Enum):
    FORALL = "forall"
    EXISTS = "exists"

@dataclass(frozen=True)
class Constraint:
    run_f: Callable[[BoundVar], bool]
    name: str = ""
    def __call__(self, bound: BoundVar)->bool:
        return self.run_f(bound)
    def __and__(self, other: Constraint) -> Constraint:
        return Constraint(
            run_f=lambda bound: self(bound) and other(bound),
            name=f"({self.name} && {other.name})"
        )
    def __or__(self, other: Constraint) -> Constraint:
        return Constraint(
            run_f=lambda bound: self(bound) or other(bound),
            name=f"({self.name} || {other.name})"
        )
    def __xor__(self, other: Constraint) -> Constraint:
        return Constraint(
            run_f=lambda bound: self(bound) ^ other(bound),
            name=f"({self.name} ^ {other.name})"
        )
    def __invert__(self) -> Constraint:
        return Constraint(
            run_f=lambda bound: not self(bound),
            name=f"!({self.name})"
        )        

    @classmethod
    def predicate(cls, f: Callable[..., bool],*, name: Optional[str] = None, quant: Quantifier = Quantifier.FORALL)->Callable[..., Constraint]:
        def wrapper(*args: Any, **kwargs:Any) -> Constraint:
            arg_list = list(args)
            kw_list = [(k, v) for k, v in kwargs.items()]
            def run_f(bound: BoundVar) -> bool:
                # positional argument values
                pos_values = [
                    arg.get(bound) if isinstance(arg, Variable) else (arg,)
                    for arg in arg_list
                ]
                # keyword argument values
                kw_keys, kw_values = zip(*[
                    (k, v.get(bound) if isinstance(v, Variable) else (v,))
                    for k, v in kw_list
                ]) if kw_list else ([], [])

                # Cartesian product over all argument values
                all_combos = product(*pos_values, *kw_values)

                # evaluate predicate on each combination
                def eval_combo(combo):
                    pos_args = combo[:len(pos_values)]
                    kw_args = dict(zip(kw_keys, combo[len(pos_values):]))
                    return f(*pos_args, **kw_args)

                if quant is Quantifier.EXISTS:
                    return any(eval_combo(c) for c in all_combos)
                else:
                    return all(eval_combo(c) for c in all_combos)
            return cls(run_f=run_f, name = name or f.__name__)
        return wrapper

    @classmethod
    def forall(cls, f: Callable[..., bool], name: Optional[str] = None) -> Callable[..., Constraint]:
        return cls.predicate(f, name=name, quant=Quantifier.FORALL)
    
    @classmethod
    def exists(cls, f: Callable[..., bool], name: Optional[str] = None):
        return cls.predicate(f, name=name, quant=Quantifier.EXISTS)


